# Copyright (C) 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the weave-era BzrDir formats.

For interface contract tests, see tests/per_bzr_dir.
"""

import os
import sys

from ... import (branch, controldir, errors, repository, upgrade, urlutils,
                 workingtree)
from ...bzr import bzrdir
from ...bzr.tests import test_bundle
from ...osutils import getcwd
from ...tests import TestCaseWithTransport
from ...tests.test_sftp_transport import TestCaseWithSFTPServer
from .branch import BzrBranchFormat4
from .bzrdir import BzrDirFormat5, BzrDirFormat6


class TestFormat5(TestCaseWithTransport):
    """Tests specific to the version 5 bzrdir format."""

    def test_same_lockfiles_between_tree_repo_branch(self):
        # this checks that only a single lockfiles instance is created
        # for format 5 objects
        dir = BzrDirFormat5().initialize(self.get_url())

        def check_dir_components_use_same_lock(dir):
            ctrl_1 = dir.open_repository().control_files
            ctrl_2 = dir.open_branch().control_files
            ctrl_3 = dir.open_workingtree()._control_files
            self.assertTrue(ctrl_1 is ctrl_2)
            self.assertTrue(ctrl_2 is ctrl_3)
        check_dir_components_use_same_lock(dir)
        # and if we open it normally.
        dir = controldir.ControlDir.open(self.get_url())
        check_dir_components_use_same_lock(dir)

    def test_can_convert(self):
        # format 5 dirs are convertable
        dir = BzrDirFormat5().initialize(self.get_url())
        self.assertTrue(dir.can_convert_format())

    def test_needs_conversion(self):
        # format 5 dirs need a conversion if they are not the default,
        # and they aren't
        dir = BzrDirFormat5().initialize(self.get_url())
        # don't need to convert it to itself
        self.assertFalse(dir.needs_format_conversion(BzrDirFormat5()))
        # do need to convert it to the current default
        self.assertTrue(dir.needs_format_conversion(
            bzrdir.BzrDirFormat.get_default_format()))


class TestFormat6(TestCaseWithTransport):
    """Tests specific to the version 6 bzrdir format."""

    def test_same_lockfiles_between_tree_repo_branch(self):
        # this checks that only a single lockfiles instance is created
        # for format 6 objects
        dir = BzrDirFormat6().initialize(self.get_url())

        def check_dir_components_use_same_lock(dir):
            ctrl_1 = dir.open_repository().control_files
            ctrl_2 = dir.open_branch().control_files
            ctrl_3 = dir.open_workingtree()._control_files
            self.assertTrue(ctrl_1 is ctrl_2)
            self.assertTrue(ctrl_2 is ctrl_3)
        check_dir_components_use_same_lock(dir)
        # and if we open it normally.
        dir = controldir.ControlDir.open(self.get_url())
        check_dir_components_use_same_lock(dir)

    def test_can_convert(self):
        # format 6 dirs are convertable
        dir = BzrDirFormat6().initialize(self.get_url())
        self.assertTrue(dir.can_convert_format())

    def test_needs_conversion(self):
        # format 6 dirs need an conversion if they are not the default.
        dir = BzrDirFormat6().initialize(self.get_url())
        self.assertTrue(dir.needs_format_conversion(
            bzrdir.BzrDirFormat.get_default_format()))


class TestBreakLockOldBranch(TestCaseWithTransport):

    def test_break_lock_format_5_bzrdir(self):
        # break lock on a format 5 bzrdir should just return
        self.make_branch_and_tree('foo', format=BzrDirFormat5())
        out, err = self.run_bzr('break-lock foo')
        self.assertEqual('', out)
        self.assertEqual('', err)


_upgrade1_template = \
    [
        ('foo', b'new contents\n'),
        ('.bzr/',),
        ('.bzr/README',
         b'This is a Bazaar control directory.\n'
         b'Do not change any files in this directory.\n'
         b'See http://bazaar.canonical.com/ for more information about Bazaar.\n'),
        ('.bzr/branch-format', b'Bazaar-NG branch, format 0.0.4\n'),
        ('.bzr/revision-history',
         b'mbp@sourcefrog.net-20051004035611-176b16534b086b3c\n'
         b'mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd\n'),
        ('.bzr/merged-patches', b''),
        ('.bzr/pending-merged-patches', b''),
        ('.bzr/branch-name', b''),
        ('.bzr/branch-lock', b''),
        ('.bzr/pending-merges', b''),
        ('.bzr/inventory',
         b'<inventory>\n'
         b'<entry file_id="foo-20051004035605-91e788d1875603ae" kind="file" name="foo" />\n'
         b'</inventory>\n'),
        ('.bzr/stat-cache',
         b'### bzr hashcache v5\n'
         b'foo// be9f309239729f69a6309e970ef24941d31e042c 13 1128398176 1128398176 303464 770\n'),
        ('.bzr/text-store/',),
        ('.bzr/text-store/foo-20051004035611-1591048e9dc7c2d4.gz',
         b'\x1f\x8b\x08\x00[\xfdAC\x02\xff\xcb\xcc\xcb,\xc9L\xccQH\xce\xcf+I\xcd+)\xe6\x02\x00\xdd\xcc\xf90\x11\x00\x00\x00'),
        ('.bzr/text-store/foo-20051004035756-4081373d897c3453.gz',
         b'\x1f\x8b\x08\x00\xc4\xfdAC\x02\xff\xcbK-WH\xce\xcf+I\xcd+)\xe6\x02\x00g\xc3\xdf\xc9\r\x00\x00\x00'),
        ('.bzr/inventory-store/',),
        ('.bzr/inventory-store/mbp@sourcefrog.net-20051004035611-176b16534b086b3c.gz',
         b'\x1f\x8b\x08\x00[\xfdAC\x02\xffm\x8f\xcd\n\xc20\x10\x84\xef>E\xc8\xbdt7?M\x02\xad\xaf"\xa1\x99`P[\xa8E\xacOo\x14\x05\x0f\xdef\xe1\xfbv\x98\xbeL7L\xeb\xbcl\xfb]_\xc3\xb2\x89\\\xce8\x944\xc8<\xcf\x8d"\xb2LdH\xdb\x8el\x13\x18\xce\xfb\xc4\xde\xd5SGHq*\xd3\x0b\xad\x8e\x14S\xbc\xe0\xadI\xb1\xe2\xbe\xfe}\xc2\xdc\xb0\rL\xc6#\xa4\xd1\x8d*\x99\x0f}=F\x1e$8G\x9d\xa0\x02\xa1rP9\x01c`FV\xda1qg\x98"\x02}\xa5\xf2\xa8\x95\xec\xa4h\xeb\x80\xf6g\xcd\x13\xb3\x01\xcc\x98\xda\x00\x00\x00'),
        ('.bzr/inventory-store/mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd.gz',
         b'\x1f\x8b\x08\x00\xc4\xfdAC\x02\xffm\x8f\xc1\n\xc20\x10D\xef~E\xc8\xbd\xb8\x9bM\x9a,\xb4\xfe\x8a\xc4f\x83Am\xa1\x16\xb1~\xbdQ\x14<x\x9b\x81y3LW\xc6\x9b\x8c\xcb4\xaf\xbbMW\xc5\xbc\xaa\\\xce\xb2/\xa9\xd7y\x9a\x1a\x03\xe0\x10\xc0\x02\xb9\x16\\\xc3(>\x84\x84\xc1WKQ\xb4:\x95\xf1\x15\xad\x8cVc\xbc\xc8\x1b\xd3j\x91\xfb\xf2\xaf\xa4r\x8d\x85\x80\xe4)\x05\xf6\x03YG\x9f\xf4\xf5\x18\xb1\xd7\x07\xe1L\xc0\x86\xd8\x1b\xce-\xc7\xb6:a\x0f\x92\x8de\x8b\x89P\xc0\x9a\xe1\x0b\x95G\x9d\xc4\xda\xb1\xad\x07\xb6?o\x9e\xb5\xff\xf0\xf9\xda\x00\x00\x00'),
        ('.bzr/revision-store/',),
        ('.bzr/revision-store/mbp@sourcefrog.net-20051004035611-176b16534b086b3c.gz',
         b'\x1f\x8b\x08\x00[\xfdAC\x02\xff\x9d\x8eKj\xc30\x14E\xe7^\x85\xd0 \xb3$\xefI\xd1\x8f\xd8\xa6\x1b(t\x07E?\xbb\x82H\n\xb2\x1ahW\xdfB1\x14:\xeb\xf4r\xee\xbdgl\xf1\x91\xb6T\x0b\xf15\xe7\xd4{l\x13}\xb6\xad\xa7B^j\xbd\x91\xc3\xad_\xb3\xbb?m\xf5\xbd\xf9\xb8\xb4\xba\x9eJ\xec\x87\xb5_)I\xe5\x11K\xaf\xed\xe35\x85\x89\xfe\xa5\x8e\x0c@ \xc0\x05\xb8\x90\x88GT\xd2\xa1\x14\xfc\xe2@K\xc7\xfd\xef\x85\xed\xcd\xe2D\x95\x8d\x1a\xa47<\x02c2\xb0 \xbc\xd0\x8ay\xa3\xbcp\x8a\x83\x12A3\xb7XJv\xef\x7f_\xf7\x94\xe3\xd6m\xbeO\x14\x91in4*<\x812\x88\xc60\xfc\x01>k\x89\x13\xe5\x12\x00\xe8<\x8c\xdf\x8d\xcd\xaeq\xb6!\x90\xa5\xd6\xf1\xbc\x07\xc3x\xde\x85\xe6\xe1\x0b\xc8\x8a\x98\x03T\x01\x00\x00'),
        ('.bzr/revision-store/mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd.gz',
         b'\x1f\x8b\x08\x00\xc4\xfdAC\x02\xff\x9d\x90Kj\x031\x0c\x86\xf79\xc5\xe0Ev\xe9\xc8o\x9b\xcc\x84^\xa0\xd0\x1b\x14\xbf&5d\xec`\xbb\x81\xf6\xf45\x84\xa4\x81\xaeZ\xa1\x85\x84^\xdf\xaf\xa9\x84K\xac1\xa7\xc1\xe5u\x8d\xad\x852\xa3\x17SZL\xc3k\xce\xa7a{j\xfb\xd5\x9e\x9fk\xfe(.,%\x1f\x9fRh\xdbc\xdb\xa3!\xa6KH-\x97\xcf\xb7\xe8g\xf4\xbbkG\x008\x06`@\xb9\xe4bG(_\x88\x95\xde\xf9n\xca\xfb\xc7\r\xf5\xdd\xe0\x19\xa9\x85)\x81\xf5"\xbd\x04j\xb8\x02b\xa8W\\\x0b\xc9\x14\xf4\xbc\xbb\xd7\xd6H4\xdc\xb8\xff}\xba\xc55\xd4f\xd6\xf3\x8c0&\x8ajE\xa4x\xe2@\xa5\xa6\x9a\xf3k\xc3WNaFT\x00\x00:l\xa6>Q\xcd1\x1cjp9\xf9;\xc34\xde\n\x9b\xe9lJWT{t\',a\xf9\x0b\xae\xc0x\x87\xa5\xb0Xp\xca,(a\xa9{\xd0{}\xd4\x12\x04(\xc5\xbb$\xc5$V\xceaI\x19\x01\xa2\x1dh\xed\x82d\x8c.\xccr@\xc3\xd8Q\xc6\x1f\xaa\xf1\xb6\xe8\xb0\xf9\x06QR\r\xf9\xfc\x01\x00\x00')]


_ghost_template = [
    ('./foo',
        b'hello\n'
     ),
    ('./.bzr/', ),
    ('./.bzr/README',
     b'This is a Bazaar control directory.\n'
     b'Do not change any files in this directory.\n'
     b'See http://bazaar.canonical.com/ for more information about Bazaar.\n'
     ),
    ('./.bzr/branch-format',
        b'Bazaar-NG branch, format 0.0.4\n'
     ),
    ('./.bzr/branch-lock',
        b''
     ),
    ('./.bzr/branch-name',
        b''
     ),
    ('./.bzr/inventory',
        b'<inventory>\n'
        b'<entry file_id="foo-20051004104918-0379cb7c76354cde" kind="file" name="foo" />\n'
        b'</inventory>\n'
     ),
    ('./.bzr/merged-patches',
        b''
     ),
    ('./.bzr/pending-merged-patches',
        b''
     ),
    ('./.bzr/pending-merges',
        b''
     ),
    ('./.bzr/revision-history',
        b'mbp@sourcefrog.net-20051004104921-a98be2278dd30b7b\n'
        b'mbp@sourcefrog.net-20051004104937-c9b7a7bfcc0bb22d\n'
     ),
    ('./.bzr/stat-cache',
        b'### bzr hashcache v5\n'
        b'foo// f572d396fae9206628714fb2ce00f72e94f2258f 6 1128422956 1128422956 306900 770\n'
     ),
    ('./.bzr/text-store/', ),
    ('./.bzr/text-store/foo-20051004104921-8de8118a71be45ba.gz',
        b'\x1f\x8b\x08\x081^BC\x00\x03foo-20051004104921-8de8118a71be45ba\x00\xcbH\xcd\xc9\xc9\xe7\x02\x00 0:6\x06\x00\x00\x00'
     ),
    ('./.bzr/inventory-store/', ),
    ('./.bzr/inventory-store/mbp@sourcefrog.net-20051004104921-a98be2278dd30b7b.gz',
        b'\x1f\x8b\x08\x081^BC\x00\x03mbp@sourcefrog.net-20051004104921-a98be2278dd30b7b\x00m\x8f\xcb\n'
        b'\xc20\x10E\xf7~E\xc8\xbe83\xcd\x13\xaa\xbf"yL0\xa8-\xd4"\xd6\xaf7\x8a\x82\x0bw\xb38\xe7\xde;C\x1do<.\xd3\xbc\xee7C;\xe6U\x94z\xe6C\xcd;Y\xa6\xa9#\x00\x8d\x00\n'
        b'Ayt\x1d\xf4\xd6\xa7h\x935\xbdV)\xb3\x14\xa7:\xbe\xd0\xe6H1\x86\x0b\xbf5)\x16\xbe/\x7fC\x08;\x97\xd9!\xba`1\xb2\xd21|\xe8\xeb1`\xe3\xb5\xa5\xdc{S\x02{\x02c\xc8YT%Rb\x80b\x89\xbd*D\xda\x95\xafT\x1f\xad\xd2H\xb1m\xfb\xb7?\xcf<\x01W}\xb5\x8b\xd9\x00\x00\x00'
     ),
    ('./.bzr/inventory-store/mbp@sourcefrog.net-20051004104937-c9b7a7bfcc0bb22d.gz',
        b'\x1f\x8b\x08\x08A^BC\x00\x03mbp@sourcefrog.net-20051004104937-c9b7a7bfcc0bb22d\x00m\x8f\xcb\n'
        b'\xc20\x10E\xf7~E\xc8\xbe83\xcd\x13\xaa\xbf"yL0\xa8-\xd4"\xd6\xaf7\x8a\x82\x0bw\xb38\xe7\xde;C\x1do<.\xd3\xbc\xee7C;\xe6U\x94z\xe6C\xcd;Y\xa6\xa9#\x00\x8d\x00\n'
        b'Ayt\x1d\xf4\xd6\xa7h\x935\xbdV)\xb3\x14\xa7:\xbe\xd0\xe6H1\x86\x0b\xbf5)\x16\xbe/\x7fC\x08;\x97\xd9!\xba`1\xb2\xd21|\xe8\xeb1`\xe3\xb5\xa5\xdc{S\x02{\x02c\xc8YT%Rb\x80b\x89\xbd*D\xda\x95\xafT\x1f\xad\xd2H\xb1m\xfb\xb7?\xcf<\x01W}\xb5\x8b\xd9\x00\x00\x00'
     ),
    ('./.bzr/revision-store/', ),
    ('./.bzr/revision-store/mbp@sourcefrog.net-20051004104921-a98be2278dd30b7b.gz',
        b'\x1f\x8b\x08\x081^BC\x00\x03mbp@sourcefrog.net-20051004104921-a98be2278dd30b7b\x00\x9d\x8eMj\xc30\x14\x84\xf7>\x85\xd0"\xbb$\xef\xc9\xb6,\x11\xdb\xf4\x02\x85\xde\xa0\xe8\xe7\xd9\x11\xc4R\x90\xd4@{\xfa\x06\x8a\xa1\xd0]\x97\x03\xdf\xcc|c\xa6G(!E\xe6\xd2\xb6\x85Z)O\xfc\xd5\xe4\x1a"{K\xe9\xc6\x0e\xb7z\xd9\xec\xfd\xa5\xa4\x8f\xech\xc9i=E\xaa\x87\xb5^8\x0b\xf1A\xb1\xa6\xfc\xf9\x1e\xfc\xc4\xffRG\x01\xd0#@\x87\xd0i\x81G\xa3\x95%!\x06\xe5}\x0bv\xb0\xbf\x17\xca\xd5\xe0\xc4-\xa0\xb1\x8b\xb6`\xc0I\xa4\xc5\xf4\x9el\xef\x95v [\x94\xcf\x8e\xd5\xcay\xe4l\xf7\xfe\xf7u\r'
        b'\x1b\x95j\xb6\xfb\xc4\x11\x85\xea\x84\xd0\x12O\x03t\x83D\xad\xc4\x0f\xf0\x95"M\xbc\x95\x00\xc0\xe7f|6\x8aYi^B.u<\xef\xb1\x19\xcf\xbb\xce\xdc|\x038=\xc7\xe6R\x01\x00\x00'
     ),
    ('./.bzr/revision-store/mbp@sourcefrog.net-20051004104937-c9b7a7bfcc0bb22d.gz',
        b'\x1f\x8b\x08\x08A^BC\x00\x03mbp@sourcefrog.net-20051004104937-c9b7a7bfcc0bb22d\x00\x9d\x90\xc1j\xc30\x0c\x86\xef}\n'
        b"\xe3Coie'\xb1c\x9a\x94\xbe\xc0`o0,[N\x03M\\\x1c\xafe{\xfae\x94n\x85\xc1`;Y\x88O\xd2\xff\xb9Mt\x19\xe6!N\xcc\xc5q\x1cr\xa6\xd4\xf1'\x9b\xf20\xb1\xe7\x18Ol}\xca\xbb\x11\xcf\x879\xbe&G!\xc5~3Q^\xf7y\xc7\xd90]h\xca1\xbd\xbd\x0c\xbe\xe3?\xa9B\x02\xd4\x02\xa0\x12P\x99R\x17\xce\xa0\xb6\x1a\x83s\x80(\xa5\x7f\xdc0\x1f\xad\xe88\x82\xb0\x18\x0c\x82\x05\xa7\x04\x05[{\xc2\xda7\xc6\x81*\x85B\x8dh\x1a\xe7\x05g\xf7\xdc\xff>\x9d\x87\x91\xe6l\xc7s\xc7\x85\x90M%\xa5\xd1z#\x85\xa8\x9b\x1a\xaa\xfa\x06\xbc\xc7\x89:^*\x00\xe0\xfbU\xbbL\xcc\xb6\xa7\xfdH\xa9'\x16\x03\xeb\x8fq\xce\xed\xf6\xde_\xb5g\x9b\x16\xa1y\xa9\xbe\x02&\n"
        b'\x7fJ+EaM\x83$\xa5n\xbc/a\x91~\xd0\xbd\xfd\x135\n'
        b'\xd0\x9a`\x0c*W\x1aR\xc1\x94du\x08(\t\xb0\x91\xdeZ\xa3\x9cU\x9cm\x7f\x8dr\x1d\x10Ot\xb8\xc6\xcf\xa7\x907|\xfb-\xb1\xbd\xd3\xfb\xd5\x07\xeeD\xee\x08*\x02\x00\x00'
     ),
]

_upgrade_dir_template = [
    ('./.bzr/', ),
    ('./.bzr/README',
     b'This is a Bazaar control directory.\n'
     b'Do not change any files in this directory.\n'
     b'See http://bazaar.canonical.com/ for more information about Bazaar.\n'
     ),
    ('./.bzr/branch-format',
        b'Bazaar-NG branch, format 0.0.4\n'
     ),
    ('./.bzr/branch-lock',
        b''
     ),
    ('./.bzr/branch-name',
        b''
     ),
    ('./.bzr/inventory',
        b'<inventory>\n'
        b'<entry file_id="dir-20051005095101-da1441ea3fa6917a" kind="directory" name="dir" />\n'
        b'</inventory>\n'
     ),
    ('./.bzr/merged-patches',
        b''
     ),
    ('./.bzr/pending-merged-patches',
        b''
     ),
    ('./.bzr/pending-merges',
        b''
     ),
    ('./.bzr/revision-history',
        b'robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e\n'
     ),
    ('./.bzr/stat-cache',
        b'### bzr hashcache v5\n'
     ),
    ('./.bzr/text-store/', ),
    ('./.bzr/inventory-store/', ),
    ('./.bzr/inventory-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
        b'\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xb3\xc9\xcc+K\xcd+\xc9/\xaa\xb4\xe3\xb2\x012\x8a*\x15\xd22sR\xe33Sl\x95R2\x8bt\x8d\x0c\x0cL\r'
        b"\x81\xd8\xc0\x12H\x19\xea\xa6$\x1a\x9a\x98\x18\xa6&\x1a\xa7%\x9aY\x1a\x9a'*)dg\xe6A\x94\xa6&\x83LQR\xc8K\xccM\x05\x0b()\xe8\x03\xcd\xd4G\xb2\x00\x00\xc2<\x94\xb1m\x00\x00\x00"
     ),
    ('./.bzr/revision-store/', ),
    ('./.bzr/revision-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
        b'\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xa5OKj\xc30\x14\xdc\xfb\x14B\x8b\xec\x92<I\xd6\xc7\xc42\x85\xde\xa0\x17(\xb6\xf4\x9c\n'
        b'l\xa9H"\x90\x9c\xbe\xa6\xa9\xa1\x9b\xae\xbax\x0c\xcc\xe71\xd3g\xbc\x85\x12R$.\xadk\xa8\x15\xb3\xa5oi\xc2\\\xc9kZ\x96\x10\x0b9,\xf5\x92\xbf)\xf7\xf2\x83O\xe5\x14\xb1\x1e\xae\xf5BI\x887\x8c5\xe5\xfb{\xf0\x96\xfei>r\x00\xc9\xb6\x83n\x03sT\xa0\xe4<y\x83\xda\x1b\xc54\xfe~T>Ff\xe9\xcc:\xdd\x8e\xa6E\xc7@\xa2\x82I\xaaNL\xbas\\313)\x00\xb9\xe6\xe0(\xd9\x87\xfc\xb7A\r'
        b"+\x96:\xae\x9f\x962\xc6\x8d\x04i\x949\x01\x97R\xb7\x1d\x17O\xc3#E\xb4T(\x00\xa0C\xd3o\x892^q\x18\xbd'>\xe4\xfe\xbc\x13M\x7f\xde{\r"
        b'\xcd\x17\x85\xea\xba\x03l\x01\x00\x00'
     ),
    ('./dir/', ),
]


class TestUpgrade(TestCaseWithTransport):

    def test_upgrade_v6_to_meta_no_workingtree(self):
        # Some format6 branches do not have checkout files. Upgrading
        # such a branch to metadir must not setup a working tree.
        self.build_tree_contents(_upgrade1_template)
        upgrade.upgrade('.', BzrDirFormat6())
        t = self.get_transport('.')
        t.delete('.bzr/pending-merges')
        t.delete('.bzr/inventory')
        self.assertFalse(t.has('.bzr/stat-cache'))
        t.delete_tree('backup.bzr.~1~')
        # At this point, we have a format6 branch without checkout files.
        upgrade.upgrade('.', bzrdir.BzrDirMetaFormat1())
        # The upgrade should not have set up a working tree.
        control = controldir.ControlDir.open('.')
        self.assertFalse(control.has_workingtree())
        # We have covered the scope of this test, we may as well check that
        # upgrade has not eaten our data, even if it's a bit redundant with
        # other tests.
        self.assertIsInstance(control._format, bzrdir.BzrDirMetaFormat1)
        b = control.open_branch()
        self.addCleanup(b.lock_read().unlock)
        self.assertEqual(b._revision_history(),
                         [b'mbp@sourcefrog.net-20051004035611-176b16534b086b3c',
                          b'mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd'])

    def test_upgrade_simple(self):
        """Upgrade simple v0.0.4 format to latest format"""
        eq = self.assertEqual
        self.build_tree_contents(_upgrade1_template)
        upgrade.upgrade('.')
        control = controldir.ControlDir.open('.')
        b = control.open_branch()
        # tsk, peeking under the covers.
        self.assertIsInstance(
            control._format,
            bzrdir.BzrDirFormat.get_default_format().__class__)
        self.addCleanup(b.lock_read().unlock)
        rh = b._revision_history()
        eq(rh,
           [b'mbp@sourcefrog.net-20051004035611-176b16534b086b3c',
            b'mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd'])
        rt = b.repository.revision_tree(rh[0])
        foo_id = b'foo-20051004035605-91e788d1875603ae'
        with rt.lock_read():
            eq(rt.get_file_text('foo'), b'initial contents\n')
        rt = b.repository.revision_tree(rh[1])
        with rt.lock_read():
            eq(rt.get_file_text('foo'), b'new contents\n')
        # check a backup was made:
        backup_dir = 'backup.bzr.~1~'
        t = self.get_transport('.')
        t.stat(backup_dir)
        t.stat(backup_dir + '/README')
        t.stat(backup_dir + '/branch-format')
        t.stat(backup_dir + '/revision-history')
        t.stat(backup_dir + '/merged-patches')
        t.stat(backup_dir + '/pending-merged-patches')
        t.stat(backup_dir + '/pending-merges')
        t.stat(backup_dir + '/branch-name')
        t.stat(backup_dir + '/branch-lock')
        t.stat(backup_dir + '/inventory')
        t.stat(backup_dir + '/stat-cache')
        t.stat(backup_dir + '/text-store')
        t.stat(backup_dir + '/text-store/foo-20051004035611-1591048e9dc7c2d4.gz')
        t.stat(backup_dir + '/text-store/foo-20051004035756-4081373d897c3453.gz')
        t.stat(backup_dir + '/inventory-store/')
        t.stat(
            backup_dir + '/inventory-store/mbp@sourcefrog.net-20051004035611-176b16534b086b3c.gz')
        t.stat(
            backup_dir + '/inventory-store/mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd.gz')
        t.stat(backup_dir + '/revision-store/')
        t.stat(
            backup_dir + '/revision-store/mbp@sourcefrog.net-20051004035611-176b16534b086b3c.gz')
        t.stat(
            backup_dir + '/revision-store/mbp@sourcefrog.net-20051004035756-235f2b7dcdddd8dd.gz')

    def test_upgrade_with_ghosts(self):
        """Upgrade v0.0.4 tree containing ghost references.

        That is, some of the parents of revisions mentioned in the branch
        aren't present in the branch's storage.

        This shouldn't normally happen in branches created entirely in
        bzr, but can happen in branches imported from baz and arch, or from
        other systems, where the importer knows about a revision but not
        its contents."""
        eq = self.assertEqual
        self.build_tree_contents(_ghost_template)
        upgrade.upgrade('.')
        b = branch.Branch.open('.')
        self.addCleanup(b.lock_read().unlock)
        revision_id = b._revision_history()[1]
        rev = b.repository.get_revision(revision_id)
        eq(len(rev.parent_ids), 2)
        eq(rev.parent_ids[1], b'wibble@wobble-2')

    def test_upgrade_makes_dir_weaves(self):
        self.build_tree_contents(_upgrade_dir_template)
        old_repodir = controldir.ControlDir.open_unsupported('.')
        old_repo_format = old_repodir.open_repository()._format
        upgrade.upgrade('.')
        # this is the path to the literal file. As format changes
        # occur it needs to be updated. FIXME: ask the store for the
        # path.
        repo = repository.Repository.open('.')
        # it should have changed the format
        self.assertNotEqual(old_repo_format.__class__, repo._format.__class__)
        # and we should be able to read the names for the file id
        # 'dir-20051005095101-da1441ea3fa6917a'
        repo.lock_read()
        self.addCleanup(repo.unlock)
        text_keys = repo.texts.keys()
        dir_keys = [key for key in text_keys if key[0] ==
                    b'dir-20051005095101-da1441ea3fa6917a']
        self.assertNotEqual([], dir_keys)

    def test_upgrade_to_meta_sets_workingtree_last_revision(self):
        self.build_tree_contents(_upgrade_dir_template)
        upgrade.upgrade('.', bzrdir.BzrDirMetaFormat1())
        tree = workingtree.WorkingTree.open('.')
        self.addCleanup(tree.lock_read().unlock)
        self.assertEqual([tree.branch._revision_history()[-1]],
                         tree.get_parent_ids())


class SFTPBranchTest(TestCaseWithSFTPServer):
    """Test some stuff when accessing a bzr Branch over sftp"""

    def test_lock_file(self):
        # old format branches use a special lock file on sftp.
        b = self.make_branch('', format=BzrDirFormat6())
        b = branch.Branch.open(self.get_url())
        self.assertPathExists('.bzr/')
        self.assertPathExists('.bzr/branch-format')
        self.assertPathExists('.bzr/branch-lock')

        self.assertPathDoesNotExist('.bzr/branch-lock.write-lock')
        b.lock_write()
        self.assertPathExists('.bzr/branch-lock.write-lock')
        b.unlock()
        self.assertPathDoesNotExist('.bzr/branch-lock.write-lock')


class TestInfo(TestCaseWithTransport):

    def test_info_locking_oslocks(self):
        if sys.platform == "win32":
            self.skip("don't use oslocks on win32 in unix manner")
        # This test tests old (all-in-one, OS lock using) behaviour which
        # simply cannot work on windows (and is indeed why we changed our
        # design. As such, don't try to remove the thisFailsStrictLockCheck
        # call here.
        self.thisFailsStrictLockCheck()

        tree = self.make_branch_and_tree('branch',
                                         format=BzrDirFormat6())

        # Test all permutations of locking the working tree, branch and repository
        # XXX: Well not yet, as we can't query oslocks yet. Currently, it's
        # implemented by raising NotImplementedError and get_physical_lock_status()
        # always returns false. This makes bzr info hide the lock status.  (Olaf)
        # W B R

        # U U U
        out, err = self.run_bzr('info -v branch')
        self.assertEqualDiff(
            """Standalone tree (format: weave)
Location:
  branch root: {}

Format:
       control: All-in-one format 6
  working tree: Working tree format 2
        branch: Branch format 4
    repository: {}

In the working tree:
         0 unchanged
         0 modified
         0 added
         0 removed
         0 renamed
         0 copied
         0 unknown
         0 ignored
         0 versioned subdirectories

Branch history:
         0 revisions

Repository:
         0 revisions
""".format('branch', tree.branch.repository._format.get_format_description(),
       ), out)
        self.assertEqual('', err)
        # L L L
        tree.lock_write()
        out, err = self.run_bzr('info -v branch')
        self.assertEqualDiff(
            """Standalone tree (format: weave)
Location:
  branch root: {}

Format:
       control: All-in-one format 6
  working tree: Working tree format 2
        branch: Branch format 4
    repository: {}

In the working tree:
         0 unchanged
         0 modified
         0 added
         0 removed
         0 renamed
         0 copied
         0 unknown
         0 ignored
         0 versioned subdirectories

Branch history:
         0 revisions

Repository:
         0 revisions
""".format('branch', tree.branch.repository._format.get_format_description(),
       ), out)
        self.assertEqual('', err)
        tree.unlock()


class TestBranchFormat4(TestCaseWithTransport):
    """Tests specific to branch format 4"""

    def test_no_metadir_support(self):
        url = self.get_url()
        bdir = bzrdir.BzrDirMetaFormat1().initialize(url)
        bdir.create_repository()
        self.assertRaises(errors.IncompatibleFormat,
                          BzrBranchFormat4().initialize, bdir)

    def test_supports_bzrdir_6(self):
        url = self.get_url()
        bdir = BzrDirFormat6().initialize(url)
        bdir.create_repository()
        BzrBranchFormat4().initialize(bdir)


class TestBoundBranch(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.build_tree(['master/', 'child/'])
        self.make_branch_and_tree('master')
        self.make_branch_and_tree('child',
                                  format=controldir.format_registry.make_controldir('weave'))
        os.chdir('child')

    def test_bind_format_6_bzrdir(self):
        # bind on a format 6 bzrdir should error
        out, err = self.run_bzr('bind ../master', retcode=3)
        self.assertEqual('', out)
        # TODO: jam 20060427 Probably something like this really should
        #       print out the actual path, rather than the URL
        cwd = urlutils.local_path_to_url(getcwd())
        self.assertEqual(
            'brz: ERROR: Branch at %s/ does not support binding.\n' % cwd, err)

    def test_unbind_format_6_bzrdir(self):
        # bind on a format 6 bzrdir should error
        out, err = self.run_bzr('unbind', retcode=3)
        self.assertEqual('', out)
        cwd = urlutils.local_path_to_url(getcwd())
        self.assertEqual('brz: ERROR: To use this feature you must '
                         'upgrade your branch at %s/.\n' % cwd, err)


class TestInit(TestCaseWithTransport):

    def test_init_weave(self):
        # --format=weave should be accepted to allow interoperation with
        # old releases when desired.
        out, err = self.run_bzr('init --format=weave')
        self.assertEqual("""Created a standalone tree (format: weave)\n""",
                         out)
        self.assertEqual('', err)


class V4WeaveBundleTester(test_bundle.V4BundleTester):

    def bzrdir_format(self):
        return 'metaweave'



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os

from breezy import trace
from breezy.rename_map import RenameMap
from breezy.tests import TestCaseWithTransport


def myhash(val):
    """This the hash used by RenameMap."""
    return hash(val) % (1024 * 1024 * 10)


class TestRenameMap(TestCaseWithTransport):

    a_lines = b'a\nb\nc\n'.splitlines(True)
    b_lines = b'b\nc\nd\n'.splitlines(True)

    def test_add_edge_hashes(self):
        rn = RenameMap(None)
        rn.add_edge_hashes(self.a_lines, 'a')
        self.assertEqual({'a'}, rn.edge_hashes[myhash(('a\n', 'b\n'))])
        self.assertEqual({'a'}, rn.edge_hashes[myhash(('b\n', 'c\n'))])
        self.assertIs(None, rn.edge_hashes.get(myhash(('c\n', 'd\n'))))

    def test_add_file_edge_hashes(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/a', b''.join(self.a_lines))])
        tree.add('a', ids=b'a')
        rn = RenameMap(tree)
        rn.add_file_edge_hashes(tree, [b'a'])
        self.assertEqual({b'a'}, rn.edge_hashes[myhash(('a\n', 'b\n'))])
        self.assertEqual({b'a'}, rn.edge_hashes[myhash(('b\n', 'c\n'))])
        self.assertIs(None, rn.edge_hashes.get(myhash(('c\n', 'd\n'))))

    def test_hitcounts(self):
        rn = RenameMap(None)
        rn.add_edge_hashes(self.a_lines, 'a')
        rn.add_edge_hashes(self.b_lines, 'b')
        self.assertEqual({'a': 2.5, 'b': 0.5}, rn.hitcounts(self.a_lines))
        self.assertEqual({'a': 1}, rn.hitcounts(self.a_lines[:-1]))
        self.assertEqual({'b': 2.5, 'a': 0.5}, rn.hitcounts(self.b_lines))

    def test_file_match(self):
        tree = self.make_branch_and_tree('tree')
        rn = RenameMap(tree)
        rn.add_edge_hashes(self.a_lines, 'aid')
        rn.add_edge_hashes(self.b_lines, 'bid')
        self.build_tree_contents([('tree/a', b''.join(self.a_lines))])
        self.build_tree_contents([('tree/b', b''.join(self.b_lines))])
        self.assertEqual({'a': 'aid', 'b': 'bid'},
                         rn.file_match(['a', 'b']))

    def test_file_match_no_dups(self):
        tree = self.make_branch_and_tree('tree')
        rn = RenameMap(tree)
        rn.add_edge_hashes(self.a_lines, 'aid')
        self.build_tree_contents([('tree/a', b''.join(self.a_lines))])
        self.build_tree_contents([('tree/b', b''.join(self.b_lines))])
        self.build_tree_contents([('tree/c', b''.join(self.b_lines))])
        self.assertEqual({'a': 'aid'},
                         rn.file_match(['a', 'b', 'c']))

    def test_match_directories(self):
        tree = self.make_branch_and_tree('tree')
        rn = RenameMap(tree)
        required_parents = rn.get_required_parents({
            'path1': 'a',
            'path2/tr': 'b',
            'path3/path4/path5': 'c',
        })
        self.assertEqual(
            {'path2': {'b'}, 'path3/path4': {'c'}, 'path3': set()},
            required_parents)

    def test_find_directory_renames(self):
        tree = self.make_branch_and_tree('tree')
        rn = RenameMap(tree)
        matches = {
            'path1': 'a',
            'path3/path4/path5': 'c',
        }
        required_parents = {
            'path2': {'b'},
            'path3/path4': {'c'},
            'path3': set()}
        missing_parents = {
            'path2-id': {'b'},
            'path4-id': {'c'},
            'path3-id': {'path4-id'}}
        matches = rn.match_parents(required_parents, missing_parents)
        self.assertEqual({'path3/path4': 'path4-id', 'path2': 'path2-id'},
                         matches)

    def test_guess_renames(self):
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['tree/file'])
        tree.add('file', ids=b'file-id')
        tree.commit('Added file')
        os.rename('tree/file', 'tree/file2')
        RenameMap.guess_renames(tree.basis_tree(), tree)
        self.assertEqual('file2', tree.id2path(b'file-id'))

    def test_guess_renames_handles_directories(self):
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['tree/dir/', 'tree/dir/file'])
        tree.add(['dir', 'dir/file'], ids=[b'dir-id', b'file-id'])
        tree.commit('Added file')
        os.rename('tree/dir', 'tree/dir2')
        RenameMap.guess_renames(tree.basis_tree(), tree)
        self.assertEqual('dir2/file', tree.id2path(b'file-id'))
        self.assertEqual('dir2', tree.id2path(b'dir-id'))

    def test_guess_renames_handles_grandparent_directories(self):
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['tree/topdir/',
                         'tree/topdir/middledir/',
                         'tree/topdir/middledir/file'])
        tree.add(['topdir', 'topdir/middledir', 'topdir/middledir/file'],
                 ids=[b'topdir-id', b'middledir-id', b'file-id'])
        tree.commit('Added files.')
        os.rename('tree/topdir', 'tree/topdir2')
        RenameMap.guess_renames(tree.basis_tree(), tree)
        self.assertEqual('topdir2', tree.id2path(b'topdir-id'))

    def test_guess_renames_preserves_children(self):
        """When a directory has been moved, its children are preserved."""
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree_contents([('tree/foo/', b''),
                                  ('tree/foo/bar', b'bar'),
                                  ('tree/foo/empty', b'')])
        tree.add(['foo', 'foo/bar', 'foo/empty'],
                 ids=[b'foo-id', b'bar-id', b'empty-id'])
        tree.commit('rev1')
        os.rename('tree/foo', 'tree/baz')
        RenameMap.guess_renames(tree.basis_tree(), tree)
        self.assertEqual('baz/empty', tree.id2path(b'empty-id'))

    def test_guess_renames_dry_run(self):
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['tree/file'])
        tree.add('file', ids=b'file-id')
        tree.commit('Added file')
        os.rename('tree/file', 'tree/file2')
        RenameMap.guess_renames(tree.basis_tree(), tree, dry_run=True)
        self.assertEqual('file', tree.id2path(b'file-id'))

    @staticmethod
    def captureNotes(cmd, *args, **kwargs):
        notes = []

        def my_note(fmt, *args):
            notes.append(fmt % args)
        old_note = trace.note
        trace.note = my_note
        try:
            result = cmd(*args, **kwargs)
        finally:
            trace.note = old_note
        return notes, result

    def test_guess_renames_output(self):
        """guess_renames emits output whether dry_run is True or False."""
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['tree/file'])
        tree.add('file', ids=b'file-id')
        tree.commit('Added file')
        os.rename('tree/file', 'tree/file2')
        notes = self.captureNotes(
            RenameMap.guess_renames, tree.basis_tree(), tree,
            dry_run=True)[0]
        self.assertEqual('file => file2', ''.join(notes))
        notes = self.captureNotes(RenameMap.guess_renames, tree.basis_tree(),
                                  tree, dry_run=False)[0]
        self.assertEqual('file => file2', ''.join(notes))

    def test_guess_rename_handles_new_directories(self):
        """When a file was moved into a new directory."""
        tree = self.make_branch_and_tree('.')
        tree.lock_write()
        # self.addCleanup(tree.unlock)
        self.build_tree(['file'])
        tree.add('file', ids=b'file-id')
        tree.commit('Added file')
        os.mkdir('folder')
        os.rename('file', 'folder/file2')
        notes = self.captureNotes(
            RenameMap.guess_renames, tree.basis_tree(), tree)[0]
        self.assertEqual('file => folder/file2', ''.join(notes))

        tree.unlock()



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import controldir
from breezy.tests import TestCaseWithTransport


class TestReference(TestCaseWithTransport):

    def get_default_format(self):
        return controldir.format_registry.make_controldir('2a')

    def test_no_args_lists(self):
        tree = self.make_branch_and_tree('branch')
        branch = tree.branch
        tree.add_reference(self.make_branch_and_tree('branch/path'))
        tree.add_reference(self.make_branch_and_tree('branch/lath'))
        tree.set_reference_info('path', 'http://example.org')
        tree.set_reference_info('lath', 'http://example.org/2')
        out, err = self.run_bzr('reference', working_dir='branch')
        lines = out.splitlines()
        self.assertEqual('lath http://example.org/2', lines[0])
        self.assertEqual('path http://example.org', lines[1])

    def make_tree_with_reference(self):
        tree = self.make_branch_and_tree('tree')
        subtree = self.make_branch_and_tree('tree/newpath')
        tree.add_reference(subtree)
        tree.set_reference_info('newpath', 'http://example.org')
        tree.commit('add reference')
        return tree

    def test_uses_working_tree_location(self):
        tree = self.make_tree_with_reference()
        out, err = self.run_bzr('reference', working_dir='tree')
        self.assertContainsRe(out, 'newpath http://example.org\n')

    def test_uses_basis_tree_location(self):
        tree = self.make_tree_with_reference()
        tree.controldir.destroy_workingtree()
        out, err = self.run_bzr('reference', working_dir='tree')
        self.assertContainsRe(out, 'newpath http://example.org\n')

    def test_one_arg_displays(self):
        tree = self.make_tree_with_reference()
        out, err = self.run_bzr('reference newpath', working_dir='tree')
        self.assertEqual('newpath http://example.org\n', out)

    def test_one_arg_uses_containing_tree(self):
        tree = self.make_tree_with_reference()
        out, err = self.run_bzr('reference -d tree newpath')
        self.assertEqual('newpath http://example.org\n', out)

    def test_two_args_sets(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file'])
        tree.add('file')
        out, err = self.run_bzr('reference -d tree file http://example.org')
        location = tree.get_reference_info('file')
        self.assertEqual('http://example.org', location)
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_missing_file(self):
        tree = self.make_branch_and_tree('tree')
        out, err = self.run_bzr('reference file http://example.org',
                                working_dir='tree', retcode=3)
        self.assertEqual('brz: ERROR: file is not versioned.\n', err)

    def test_missing_file_forced(self):
        tree = self.make_branch_and_tree('tree')
        tree.add_reference(self.make_branch_and_tree('tree/file'))
        out, err = self.run_bzr(
            'reference --force-unversioned file http://example.org',
            working_dir='tree')
        location = tree.get_reference_info('file')
        self.assertEqual('http://example.org', location)
        self.assertEqual('', out)
        self.assertEqual('', err)



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for LockDir"""

import os
import time

import breezy

from .. import config, errors, lock, lockdir, osutils, tests, transport
from ..errors import (LockBreakMismatch, LockBroken, LockContention,
                      LockFailed, LockNotHeld)
from ..lockdir import LockDir, LockHeldInfo
from . import TestCase, TestCaseInTempDir, TestCaseWithTransport, features

# These tests are run on the default transport provided by the test framework
# (typically a local disk transport).  That can be changed by the --transport
# option to bzr selftest.  The required properties of the transport
# implementation are tested separately.  (The main requirement is just that
# they don't allow overwriting nonempty directories.)


class TestLockDir(TestCaseWithTransport):
    """Test LockDir operations"""

    def logging_report_function(self, fmt, *args):
        self._logged_reports.append((fmt, args))

    def setup_log_reporter(self, lock_dir):
        self._logged_reports = []
        lock_dir._report_function = self.logging_report_function

    def test_00_lock_creation(self):
        """Creation of lock file on a transport"""
        t = self.get_transport()
        lf = LockDir(t, 'test_lock')
        self.assertFalse(lf.is_held)

    def test_01_lock_repr(self):
        """Lock string representation"""
        lf = LockDir(self.get_transport(), 'test_lock')
        r = repr(lf)
        self.assertContainsRe(r, r'^LockDir\(.*/test_lock\)$')

    def test_02_unlocked_peek(self):
        lf = LockDir(self.get_transport(), 'test_lock')
        self.assertEqual(lf.peek(), None)

    def get_lock(self):
        return LockDir(self.get_transport(), 'test_lock')

    def test_unlock_after_break_raises(self):
        ld = self.get_lock()
        ld2 = self.get_lock()
        ld.create()
        ld.attempt_lock()
        ld2.force_break(ld2.peek())
        self.assertRaises(LockBroken, ld.unlock)

    def test_03_readonly_peek(self):
        lf = LockDir(self.get_readonly_transport(), 'test_lock')
        self.assertEqual(lf.peek(), None)

    def test_10_lock_uncontested(self):
        """Acquire and release a lock"""
        t = self.get_transport()
        lf = LockDir(t, 'test_lock')
        lf.create()
        lf.attempt_lock()
        try:
            self.assertTrue(lf.is_held)
        finally:
            lf.unlock()
            self.assertFalse(lf.is_held)

    def test_11_create_readonly_transport(self):
        """Fail to create lock on readonly transport"""
        t = self.get_readonly_transport()
        lf = LockDir(t, 'test_lock')
        self.assertRaises(LockFailed, lf.create)

    def test_12_lock_readonly_transport(self):
        """Fail to lock on readonly transport"""
        lf = LockDir(self.get_transport(), 'test_lock')
        lf.create()
        lf = LockDir(self.get_readonly_transport(), 'test_lock')
        self.assertRaises(LockFailed, lf.attempt_lock)

    def test_20_lock_contested(self):
        """Contention to get a lock"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        lf2 = LockDir(t, 'test_lock')
        try:
            # locking is between LockDir instances; aliases within
            # a single process are not detected
            lf2.attempt_lock()
            self.fail('Failed to detect lock collision')
        except LockContention as e:
            self.assertEqual(e.lock, lf2)
            self.assertContainsRe(str(e),
                                  r'^Could not acquire.*test_lock.*$')
        lf1.unlock()

    def test_20_lock_peek(self):
        """Peek at the state of a lock"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        self.addCleanup(lf1.unlock)
        # lock is held, should get some info on it
        info1 = lf1.peek()
        self.assertEqual(set(info1.info_dict.keys()),
                         {'user', 'nonce', 'hostname', 'pid', 'start_time'})
        # should get the same info if we look at it through a different
        # instance
        info2 = LockDir(t, 'test_lock').peek()
        self.assertEqual(info1, info2)
        # locks which are never used should be not-held
        self.assertEqual(LockDir(t, 'other_lock').peek(), None)

    def test_21_peek_readonly(self):
        """Peek over a readonly transport"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf2 = LockDir(self.get_readonly_transport(), 'test_lock')
        self.assertEqual(lf2.peek(), None)
        lf1.attempt_lock()
        self.addCleanup(lf1.unlock)
        info2 = lf2.peek()
        self.assertTrue(info2)
        self.assertEqual(info2.nonce, lf1.nonce)

    def test_30_lock_wait_fail(self):
        """Wait on a lock, then fail

        We ask to wait up to 400ms; this should fail within at most one
        second.  (Longer times are more realistic but we don't want the test
        suite to take too long, and this should do for now.)
        """
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf2 = LockDir(t, 'test_lock')
        self.setup_log_reporter(lf2)
        lf1.attempt_lock()
        try:
            before = time.time()
            self.assertRaises(LockContention, lf2.wait_lock,
                              timeout=0.4, poll=0.1)
            after = time.time()
            # it should only take about 0.4 seconds, but we allow more time in
            # case the machine is heavily loaded
            self.assertTrue(after - before <= 8.0,
                            "took %f seconds to detect lock contention" % (after - before))
        finally:
            lf1.unlock()
        self.assertEqual(1, len(self._logged_reports))
        self.assertContainsRe(self._logged_reports[0][0],
                              r'Unable to obtain lock .* held by jrandom@example\.com on .*'
                              r' \(process #\d+\), acquired .* ago\.\n'
                              r'Will continue to try until \d{2}:\d{2}:\d{2}, unless '
                              r'you press Ctrl-C.\n'
                              r'See "brz help break-lock" for more.')

    def test_31_lock_wait_easy(self):
        """Succeed when waiting on a lock with no contention.
        """
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        self.setup_log_reporter(lf1)
        try:
            before = time.time()
            lf1.wait_lock(timeout=0.4, poll=0.1)
            after = time.time()
            self.assertTrue(after - before <= 1.0)
        finally:
            lf1.unlock()
        self.assertEqual([], self._logged_reports)

    def test_40_confirm_easy(self):
        """Confirm a lock that's already held"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        self.addCleanup(lf1.unlock)
        lf1.confirm()

    def test_41_confirm_not_held(self):
        """Confirm a lock that's already held"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        self.assertRaises(LockNotHeld, lf1.confirm)

    def test_42_confirm_broken_manually(self):
        """Confirm a lock broken by hand"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        t.move('test_lock', 'lock_gone_now')
        self.assertRaises(LockBroken, lf1.confirm)
        # Clean up
        t.move('lock_gone_now', 'test_lock')
        lf1.unlock()

    def test_43_break(self):
        """Break a lock whose caller has forgotten it"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        # we incorrectly discard the lock object without unlocking it
        del lf1
        # someone else sees it's still locked
        lf2 = LockDir(t, 'test_lock')
        holder_info = lf2.peek()
        self.assertTrue(holder_info)
        lf2.force_break(holder_info)
        # now we should be able to take it
        lf2.attempt_lock()
        self.addCleanup(lf2.unlock)
        lf2.confirm()

    def test_44_break_already_released(self):
        """Lock break races with regular release"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        # someone else sees it's still locked
        lf2 = LockDir(t, 'test_lock')
        holder_info = lf2.peek()
        # in the interim the lock is released
        lf1.unlock()
        # break should succeed
        lf2.force_break(holder_info)
        # now we should be able to take it
        lf2.attempt_lock()
        self.addCleanup(lf2.unlock)
        lf2.confirm()

    def test_45_break_mismatch(self):
        """Lock break races with someone else acquiring it"""
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.attempt_lock()
        # someone else sees it's still locked
        lf2 = LockDir(t, 'test_lock')
        holder_info = lf2.peek()
        # in the interim the lock is released
        lf1.unlock()
        lf3 = LockDir(t, 'test_lock')
        lf3.attempt_lock()
        # break should now *fail*
        self.assertRaises(LockBreakMismatch, lf2.force_break,
                          holder_info)
        lf3.unlock()

    def test_46_fake_read_lock(self):
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        lf1.lock_read()
        lf1.unlock()

    def test_50_lockdir_representation(self):
        """Check the on-disk representation of LockDirs is as expected.

        There should always be a top-level directory named by the lock.
        When the lock is held, there should be a lockname/held directory
        containing an info file.
        """
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')
        lf1.create()
        self.assertTrue(t.has('test_lock'))
        lf1.lock_write()
        self.assertTrue(t.has('test_lock/held/info'))
        lf1.unlock()
        self.assertFalse(t.has('test_lock/held/info'))

    def test_break_lock(self):
        # the ui based break_lock routine should Just Work (tm)
        ld1 = self.get_lock()
        ld2 = self.get_lock()
        ld1.create()
        ld1.lock_write()
        # do this without IO redirection to ensure it doesn't prompt.
        self.assertRaises(AssertionError, ld1.break_lock)
        orig_factory = breezy.ui.ui_factory
        breezy.ui.ui_factory = breezy.ui.CannedInputUIFactory([True])
        try:
            ld2.break_lock()
            self.assertRaises(LockBroken, ld1.unlock)
        finally:
            breezy.ui.ui_factory = orig_factory

    def test_break_lock_corrupt_info(self):
        """break_lock works even if the info file is corrupt (and tells the UI
        that it is corrupt).
        """
        ld = self.get_lock()
        ld2 = self.get_lock()
        ld.create()
        ld.lock_write()
        ld.transport.put_bytes_non_atomic('test_lock/held/info', b'\0')

        class LoggingUIFactory(breezy.ui.SilentUIFactory):
            def __init__(self):
                self.prompts = []

            def get_boolean(self, prompt):
                self.prompts.append(('boolean', prompt))
                return True

        ui = LoggingUIFactory()
        self.overrideAttr(breezy.ui, 'ui_factory', ui)
        ld2.break_lock()
        self.assertLength(1, ui.prompts)
        self.assertEqual('boolean', ui.prompts[0][0])
        self.assertStartsWith(ui.prompts[0][1], 'Break (corrupt LockDir')
        self.assertRaises(LockBroken, ld.unlock)

    def test_break_lock_missing_info(self):
        """break_lock works even if the info file is missing (and tells the UI
        that it is corrupt).
        """
        ld = self.get_lock()
        ld2 = self.get_lock()
        ld.create()
        ld.lock_write()
        ld.transport.delete('test_lock/held/info')

        class LoggingUIFactory(breezy.ui.SilentUIFactory):
            def __init__(self):
                self.prompts = []

            def get_boolean(self, prompt):
                self.prompts.append(('boolean', prompt))
                return True

        ui = LoggingUIFactory()
        orig_factory = breezy.ui.ui_factory
        breezy.ui.ui_factory = ui
        try:
            ld2.break_lock()
            self.assertRaises(LockBroken, ld.unlock)
            self.assertLength(0, ui.prompts)
        finally:
            breezy.ui.ui_factory = orig_factory
        # Suppress warnings due to ld not being unlocked
        # XXX: if lock_broken hook was invoked in this case, this hack would
        # not be necessary.  - Andrew Bennetts, 2010-09-06.
        del self._lock_actions[:]

    def test_create_missing_base_directory(self):
        """If LockDir.path doesn't exist, it can be created

        Some people manually remove the entire lock/ directory trying
        to unlock a stuck repository/branch/etc. Rather than failing
        after that, just create the lock directory when needed.
        """
        t = self.get_transport()
        lf1 = LockDir(t, 'test_lock')

        lf1.create()
        self.assertTrue(t.has('test_lock'))

        t.rmdir('test_lock')
        self.assertFalse(t.has('test_lock'))

        # This will create 'test_lock' if it needs to
        lf1.lock_write()
        self.assertTrue(t.has('test_lock'))
        self.assertTrue(t.has('test_lock/held/info'))

        lf1.unlock()
        self.assertFalse(t.has('test_lock/held/info'))

    def test_display_form(self):
        ld1 = self.get_lock()
        ld1.create()
        ld1.lock_write()
        try:
            info_list = ld1.peek().to_readable_dict()
        finally:
            ld1.unlock()
        self.assertEqual(info_list['user'], 'jrandom@example.com')
        self.assertIsInstance(info_list['pid'], int)
        self.assertContainsRe(info_list['time_ago'], '^\\d+ seconds? ago$')

    def test_lock_without_email(self):
        global_config = config.GlobalStack()
        # Intentionally has no email address
        global_config.set('email', 'User Identity')
        ld1 = self.get_lock()
        ld1.create()
        ld1.lock_write()
        ld1.unlock()

    def test_lock_permission(self):
        self.requireFeature(features.not_running_as_root)
        if not osutils.supports_posix_readonly():
            raise tests.TestSkipped('Cannot induce a permission failure')
        ld1 = self.get_lock()
        lock_path = ld1.transport.local_abspath('test_lock')
        os.mkdir(lock_path)
        osutils.make_readonly(lock_path)
        self.assertRaises(errors.LockFailed, ld1.attempt_lock)

    def test_lock_by_token(self):
        ld1 = self.get_lock()
        token = ld1.lock_write()
        self.addCleanup(ld1.unlock)
        self.assertNotEqual(None, token)
        ld2 = self.get_lock()
        t2 = ld2.lock_write(token)
        self.addCleanup(ld2.unlock)
        self.assertEqual(token, t2)

    def test_lock_with_buggy_rename(self):
        # test that lock acquisition handles servers which pretend they
        # renamed correctly but that actually fail
        t = transport.get_transport_from_url(
            'brokenrename+' + self.get_url())
        ld1 = LockDir(t, 'test_lock')
        ld1.create()
        ld1.attempt_lock()
        ld2 = LockDir(t, 'test_lock')
        # we should fail to lock
        e = self.assertRaises(errors.LockContention, ld2.attempt_lock)
        # now the original caller should succeed in unlocking
        ld1.unlock()
        # and there should be nothing left over
        self.assertEqual([], t.list_dir('test_lock'))

    def test_failed_lock_leaves_no_trash(self):
        # if we fail to acquire the lock, we don't leave pending directories
        # behind -- https://bugs.launchpad.net/bzr/+bug/109169
        ld1 = self.get_lock()
        ld2 = self.get_lock()
        # should be nothing before we start
        ld1.create()
        t = self.get_transport().clone('test_lock')

        def check_dir(a):
            self.assertEqual(a, t.list_dir('.'))

        check_dir([])
        # when held, that's all we see
        ld1.attempt_lock()
        self.addCleanup(ld1.unlock)
        check_dir(['held'])
        # second guy should fail
        self.assertRaises(errors.LockContention, ld2.attempt_lock)
        # no kibble
        check_dir(['held'])

    def test_no_lockdir_info(self):
        """We can cope with empty info files."""
        # This seems like a fairly common failure case - see
        # <https://bugs.launchpad.net/bzr/+bug/185103> and all its dupes.
        # Processes are often interrupted after opening the file
        # before the actual contents are committed.
        t = self.get_transport()
        t.mkdir('test_lock')
        t.mkdir('test_lock/held')
        t.put_bytes('test_lock/held/info', b'')
        lf = LockDir(t, 'test_lock')
        info = lf.peek()
        formatted_info = info.to_readable_dict()
        self.assertEqual(
            dict(user='<unknown>', hostname='<unknown>', pid='<unknown>',
                 time_ago='(unknown)'),
            formatted_info)

    def test_corrupt_lockdir_info(self):
        """We can cope with corrupt (and thus unparseable) info files."""
        # This seems like a fairly common failure case too - see
        # <https://bugs.launchpad.net/bzr/+bug/619872> for instance.
        # In particular some systems tend to fill recently created files with
        # nul bytes after recovering from a system crash.
        t = self.get_transport()
        t.mkdir('test_lock')
        t.mkdir('test_lock/held')
        t.put_bytes('test_lock/held/info', b'\0')
        lf = LockDir(t, 'test_lock')
        self.assertRaises(errors.LockCorrupt, lf.peek)
        # Currently attempt_lock gives LockContention, but LockCorrupt would be
        # a reasonable result too.
        self.assertRaises(
            (errors.LockCorrupt, errors.LockContention), lf.attempt_lock)
        self.assertRaises(errors.LockCorrupt, lf.validate_token, 'fake token')

    def test_missing_lockdir_info(self):
        """We can cope with absent info files."""
        t = self.get_transport()
        t.mkdir('test_lock')
        t.mkdir('test_lock/held')
        lf = LockDir(t, 'test_lock')
        # In this case we expect the 'not held' result from peek, because peek
        # cannot be expected to notice that there is a 'held' directory with no
        # 'info' file.
        self.assertEqual(None, lf.peek())
        # And lock/unlock may work or give LockContention (but not any other
        # error).
        try:
            lf.attempt_lock()
        except LockContention:
            # LockContention is ok, and expected on Windows
            pass
        else:
            # no error is ok, and expected on POSIX (because POSIX allows
            # os.rename over an empty directory).
            lf.unlock()
        # Currently raises TokenMismatch, but LockCorrupt would be reasonable
        # too.
        self.assertRaises(
            (errors.TokenMismatch, errors.LockCorrupt),
            lf.validate_token, 'fake token')


class TestLockDirHooks(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self._calls = []

    def get_lock(self):
        return LockDir(self.get_transport(), 'test_lock')

    def record_hook(self, result):
        self._calls.append(result)

    def test_LockDir_acquired_success(self):
        # the LockDir.lock_acquired hook fires when a lock is acquired.
        LockDir.hooks.install_named_hook('lock_acquired',
                                         self.record_hook, 'record_hook')
        ld = self.get_lock()
        ld.create()
        self.assertEqual([], self._calls)
        result = ld.attempt_lock()
        lock_path = ld.transport.abspath(ld.path)
        self.assertEqual([lock.LockResult(lock_path, result)], self._calls)
        ld.unlock()
        self.assertEqual([lock.LockResult(lock_path, result)], self._calls)

    def test_LockDir_acquired_fail(self):
        # the LockDir.lock_acquired hook does not fire on failure.
        ld = self.get_lock()
        ld.create()
        ld2 = self.get_lock()
        ld2.attempt_lock()
        # install a lock hook now, when the disk lock is locked
        LockDir.hooks.install_named_hook('lock_acquired',
                                         self.record_hook, 'record_hook')
        self.assertRaises(errors.LockContention, ld.attempt_lock)
        self.assertEqual([], self._calls)
        ld2.unlock()
        self.assertEqual([], self._calls)

    def test_LockDir_released_success(self):
        # the LockDir.lock_released hook fires when a lock is acquired.
        LockDir.hooks.install_named_hook('lock_released',
                                         self.record_hook, 'record_hook')
        ld = self.get_lock()
        ld.create()
        self.assertEqual([], self._calls)
        result = ld.attempt_lock()
        self.assertEqual([], self._calls)
        ld.unlock()
        lock_path = ld.transport.abspath(ld.path)
        self.assertEqual([lock.LockResult(lock_path, result)], self._calls)

    def test_LockDir_released_fail(self):
        # the LockDir.lock_released hook does not fire on failure.
        ld = self.get_lock()
        ld.create()
        ld2 = self.get_lock()
        ld.attempt_lock()
        ld2.force_break(ld2.peek())
        LockDir.hooks.install_named_hook('lock_released',
                                         self.record_hook, 'record_hook')
        self.assertRaises(LockBroken, ld.unlock)
        self.assertEqual([], self._calls)

    def test_LockDir_broken_success(self):
        # the LockDir.lock_broken hook fires when a lock is broken.
        ld = self.get_lock()
        ld.create()
        ld2 = self.get_lock()
        result = ld.attempt_lock()
        LockDir.hooks.install_named_hook('lock_broken',
                                         self.record_hook, 'record_hook')
        ld2.force_break(ld2.peek())
        lock_path = ld.transport.abspath(ld.path)
        self.assertEqual([lock.LockResult(lock_path, result)], self._calls)

    def test_LockDir_broken_failure(self):
        # the LockDir.lock_broken hook does not fires when a lock is already
        # released.
        ld = self.get_lock()
        ld.create()
        ld2 = self.get_lock()
        result = ld.attempt_lock()
        holder_info = ld2.peek()
        ld.unlock()
        LockDir.hooks.install_named_hook('lock_broken',
                                         self.record_hook, 'record_hook')
        ld2.force_break(holder_info)
        lock_path = ld.transport.abspath(ld.path)
        self.assertEqual([], self._calls)


class TestLockHeldInfo(TestCaseInTempDir):
    """Can get information about the lock holder, and detect whether they're
    still alive."""

    def test_repr(self):
        info = LockHeldInfo.for_this_process(None)
        self.assertContainsRe(repr(info), r"LockHeldInfo\(.*\)")

    def test_unicode(self):
        info = LockHeldInfo.for_this_process(None)
        self.assertContainsRe(str(info),
                              r'held by .* on .* \(process #\d+\), acquired .* ago')

    def test_is_locked_by_this_process(self):
        info = LockHeldInfo.for_this_process(None)
        self.assertTrue(info.is_locked_by_this_process())

    def test_is_not_locked_by_this_process(self):
        info = LockHeldInfo.for_this_process(None)
        info.info_dict['pid'] = '123123123123123'
        self.assertFalse(info.is_locked_by_this_process())

    def test_lock_holder_live_process(self):
        """Detect that the holder (this process) is still running."""
        info = LockHeldInfo.for_this_process(None)
        self.assertFalse(info.is_lock_holder_known_dead())

    def test_lock_holder_dead_process(self):
        """Detect that the holder (this process) is still running."""
        self.overrideAttr(lockdir, 'get_host_name',
                          lambda: 'aproperhostname')
        info = LockHeldInfo.for_this_process(None)
        info.info_dict['pid'] = '123123123'
        self.assertTrue(info.is_lock_holder_known_dead())

    def test_lock_holder_other_machine(self):
        """The lock holder isn't here so we don't know if they're alive."""
        info = LockHeldInfo.for_this_process(None)
        info.info_dict['hostname'] = 'egg.example.com'
        info.info_dict['pid'] = '123123123'
        self.assertFalse(info.is_lock_holder_known_dead())

    def test_lock_holder_other_user(self):
        """Only auto-break locks held by this user."""
        info = LockHeldInfo.for_this_process(None)
        info.info_dict['user'] = 'notme@example.com'
        info.info_dict['pid'] = '123123123'
        self.assertFalse(info.is_lock_holder_known_dead())

    def test_no_good_hostname(self):
        """Correctly handle ambiguous hostnames.

        If the lock's recorded with just 'localhost' we can't really trust
        it's the same 'localhost'.  (There are quite a few of them. :-)
        So even if the process is known not to be alive, we can't say that's
        known for sure.
        """
        self.overrideAttr(lockdir, 'get_host_name',
                          lambda: 'localhost')
        info = LockHeldInfo.for_this_process(None)
        info.info_dict['pid'] = '123123123'
        self.assertFalse(info.is_lock_holder_known_dead())


class TestStaleLockDir(TestCaseWithTransport):
    """Can automatically break stale locks.

    :see: https://bugs.launchpad.net/bzr/+bug/220464
    """

    def test_auto_break_stale_lock(self):
        """Locks safely known to be stale are just cleaned up.

        This generates a warning but no other user interaction.
        """
        self.overrideAttr(lockdir, 'get_host_name',
                          lambda: 'aproperhostname')
        # Stealing dead locks is enabled by default.
        # Create a lock pretending to come from a different nonexistent
        # process on the same machine.
        l1 = LockDir(self.get_transport(), 'a',
                     extra_holder_info={'pid': '12312313'})
        token_1 = l1.attempt_lock()
        l2 = LockDir(self.get_transport(), 'a')
        token_2 = l2.attempt_lock()
        # l1 will notice its lock was stolen.
        self.assertRaises(errors.LockBroken,
                          l1.unlock)
        l2.unlock()

    def test_auto_break_stale_lock_configured_off(self):
        """Automatic breaking can be turned off"""
        l1 = LockDir(self.get_transport(), 'a',
                     extra_holder_info={'pid': '12312313'})
        # Stealing dead locks is enabled by default, so disable it.
        config.GlobalStack().set('locks.steal_dead', False)
        token_1 = l1.attempt_lock()
        self.addCleanup(l1.unlock)
        l2 = LockDir(self.get_transport(), 'a')
        # This fails now, because dead lock breaking is disabled.
        self.assertRaises(LockContention,
                          l2.attempt_lock)
        # and it's in fact not broken
        l1.confirm()



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2006-2009 Canonical Ltd

# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer@jelmer.uk>
#          John Carr <john.carr@unrouted.co.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""A GIT branch and repository format implementation for bzr."""

import os
import sys

dulwich_minimum_version = (0, 19, 11)

from .. import __version__ as breezy_version  # noqa: F401
from .. import errors as brz_errors
from .. import trace, urlutils, version_info
from ..commands import plugin_cmds
from ..controldir import ControlDirFormat, Prober, format_registry
from ..controldir import \
    network_format_registry as controldir_network_format_registry
from ..transport import (register_lazy_transport, register_transport_proto,
                         transport_server_registry)

if getattr(sys, "frozen", None):
    # allow import additional libs from ./_lib for bzr.exe only
    sys.path.append(os.path.normpath(
        os.path.join(os.path.dirname(__file__), '_lib')))


def import_dulwich():
    try:
        from dulwich import __version__ as dulwich_version
    except ModuleNotFoundError:
        raise brz_errors.DependencyNotPresent(
            "dulwich",
            "bzr-git: Please install dulwich, https://www.dulwich.io/")
    else:
        if dulwich_version < dulwich_minimum_version:
            raise brz_errors.DependencyNotPresent(
                "dulwich",
                "bzr-git: Dulwich is too old; at least %d.%d.%d is required" %
                dulwich_minimum_version)


_versions_checked = False


def lazy_check_versions():
    global _versions_checked
    if _versions_checked:
        return
    import_dulwich()
    _versions_checked = True


format_registry.register_lazy(
    'git', __name__ + ".dir", "LocalGitControlDirFormat",
    help='GIT repository.', native=False, experimental=False)

format_registry.register_lazy(
    'git-bare', __name__ + ".dir", "BareLocalGitControlDirFormat",
    help='Bare GIT repository (no working tree).', native=False,
    experimental=False)

from ..revisionspec import RevisionSpec_dwim, revspec_registry

revspec_registry.register_lazy("git:", __name__ + ".revspec",
                               "RevisionSpec_git")
RevisionSpec_dwim.append_possible_lazy_revspec(
    __name__ + ".revspec", "RevisionSpec_git")


class LocalGitProber(Prober):

    @classmethod
    def priority(klass, transport):
        return 10

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except brz_errors.InProcessTransport:
            raise brz_errors.NotBranchError(path=transport.base)
        if (external_url.startswith("http:") or
                external_url.startswith("https:")):
            # Already handled by RemoteGitProber
            raise brz_errors.NotBranchError(path=transport.base)
        if urlutils.split(transport.base)[1] == ".git":
            raise brz_errors.NotBranchError(path=transport.base)
        if not transport.has_any(['objects', '.git/objects', '.git']):
            raise brz_errors.NotBranchError(path=transport.base)
        lazy_check_versions()
        from .dir import BareLocalGitControlDirFormat, LocalGitControlDirFormat
        if transport.has_any(['.git/objects', '.git']):
            return LocalGitControlDirFormat()
        if transport.has('info') and transport.has('objects'):
            return BareLocalGitControlDirFormat()
        raise brz_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .dir import BareLocalGitControlDirFormat, LocalGitControlDirFormat
        return [BareLocalGitControlDirFormat(), LocalGitControlDirFormat()]


def user_agent_for_github():
    # GitHub requires we lie. https://github.com/dulwich/dulwich/issues/562
    return "git/Breezy/%s" % breezy_version


def is_github_url(url):
    (scheme, user, password, host, port,
     path) = urlutils.parse_url(url)
    return host in ("github.com", "gopkg.in")


class RemoteGitProber(Prober):

    @classmethod
    def priority(klass, transport):
        # This is a surprisingly good heuristic to determine whether this
        # prober is more likely to succeed than the Bazaar one.
        if 'git' in transport.base:
            return -15
        return -10

    def probe_http_transport(self, transport):
        # This function intentionally doesn't use any of the support code under
        # breezy.git, since it's called for every repository that's
        # accessed over HTTP, whether it's Git, Bzr or something else.
        # Importing Dulwich and the other support code adds unnecessray slowdowns.
        base_url = urlutils.strip_segment_parameters(transport.external_url())
        url = urlutils.URL.from_string(base_url)
        url.user = url.quoted_user = None
        url.password = url.quoted_password = None
        host = url.host
        url = urlutils.join(str(url), "info/refs") + "?service=git-upload-pack"
        headers = {"Content-Type": "application/x-git-upload-pack-request",
                   "Accept": "application/x-git-upload-pack-result",
                   }
        if is_github_url(url):
            # GitHub requires we lie.
            # https://github.com/dulwich/dulwich/issues/562
            headers["User-Agent"] = user_agent_for_github()
        resp = transport.request('GET', url, headers=headers)
        if resp.status in (404, 405):
            raise brz_errors.NotBranchError(transport.base)
        elif resp.status == 400 and resp.reason == 'no such method: info':
            # hgweb :(
            raise brz_errors.NotBranchError(transport.base)
        elif resp.status != 200:
            raise brz_errors.UnexpectedHttpStatus(
                url, resp.status, headers=resp.getheaders())

        ct = resp.getheader("Content-Type")
        if ct and ct.startswith("application/x-git"):
            from .remote import RemoteGitControlDirFormat
            return RemoteGitControlDirFormat()
        elif not ct:
            from .dir import BareLocalGitControlDirFormat
            ret = BareLocalGitControlDirFormat()
            ret._refs_text = resp.read()
            return ret
        raise brz_errors.NotBranchError(transport.base)

    def probe_transport(self, transport):
        try:
            external_url = transport.external_url()
        except brz_errors.InProcessTransport:
            raise brz_errors.NotBranchError(path=transport.base)

        if (external_url.startswith("http:") or
                external_url.startswith("https:")):
            return self.probe_http_transport(transport)

        if (not external_url.startswith("git://") and
                not external_url.startswith("git+")):
            raise brz_errors.NotBranchError(transport.base)

        # little ugly, but works
        from .remote import GitSmartTransport, RemoteGitControlDirFormat
        if isinstance(transport, GitSmartTransport):
            return RemoteGitControlDirFormat()
        raise brz_errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        from .remote import RemoteGitControlDirFormat
        return [RemoteGitControlDirFormat()]


ControlDirFormat.register_prober(LocalGitProber)
ControlDirFormat.register_prober(RemoteGitProber)

register_transport_proto(
    'git://', help="Access using the Git smart server protocol.")
register_transport_proto(
    'git+ssh://',
    help="Access using the Git smart server protocol over SSH.")

register_lazy_transport("git://", __name__ + '.remote',
                        'TCPGitSmartTransport')
register_lazy_transport("git+ssh://", __name__ + '.remote',
                        'SSHGitSmartTransport')


plugin_cmds.register_lazy("cmd_git_import", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_object", ["git-objects", "git-cat"],
                          __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_refs", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_apply", [], __name__ + ".commands")
plugin_cmds.register_lazy("cmd_git_push_pristine_tar_deltas",
                          ['git-push-pristine-tar', 'git-push-pristine'],
                          __name__ + ".commands")


def extract_git_foreign_revid(rev):
    try:
        foreign_revid = rev.foreign_revid
    except AttributeError:
        from .mapping import mapping_registry
        foreign_revid, mapping = \
            mapping_registry.parse_revision_id(rev.revision_id)
        return foreign_revid
    else:
        from .mapping import foreign_vcs_git
        if rev.mapping.vcs == foreign_vcs_git:
            return foreign_revid
        else:
            raise brz_errors.InvalidRevisionId(rev.revision_id, None)


def update_stanza(rev, stanza):
    try:
        git_commit = extract_git_foreign_revid(rev)
    except brz_errors.InvalidRevisionId:
        pass
    else:
        stanza.add("git-commit", git_commit)


from ..hooks import install_lazy_named_hook

install_lazy_named_hook(
    "breezy.version_info_formats.format_rio",
    "RioVersionInfoBuilder.hooks", "revision", update_stanza,
    "git commits")


def rewrite_instead_of(location, purpose):
    from dulwich.config import StackedConfig, iter_instead_of

    config = StackedConfig.default()

    push = (purpose != "read")

    longest_needle = ""
    updated_url = location
    for needle, replacement in iter_instead_of(config, push):
        if not location.startswith(needle):
            continue
        if len(longest_needle) < len(needle):
            longest_needle = needle
            if longest_needle == "lp:":
                # Leave the lp: prefix to the launchpad plugin, if loaded
                import breezy.plugins
                if hasattr(breezy.plugins, "launchpad"):
                    trace.warning(
                        "Ignoring insteadOf lp: in git config, because the Launchpad plugin is loaded.")
                    continue
            updated_url = replacement + location[len(needle):]
    return updated_url


from ..location import hooks as location_hooks

location_hooks.install_named_hook(
    "rewrite_location", rewrite_instead_of,
    "apply Git insteadOf / pushInsteadOf")

transport_server_registry.register_lazy(
    'git', __name__ + '.server', 'serve_git',
    'Git Smart server protocol over TCP. (default port: 9418)')

transport_server_registry.register_lazy(
    'git-receive-pack', __name__ + '.server',
    'serve_git_receive_pack',
    help='Git Smart server receive pack command. (inetd mode only)')
transport_server_registry.register_lazy(
    'git-upload-pack', __name__ + 'git.server',
    'serve_git_upload_pack',
    help='Git Smart server upload pack command. (inetd mode only)')

from ..repository import format_registry as repository_format_registry
from ..repository import \
    network_format_registry as repository_network_format_registry

repository_network_format_registry.register_lazy(
    b'git', __name__ + '.repository', 'GitRepositoryFormat')

register_extra_lazy_repository_format = getattr(repository_format_registry,
                                                "register_extra_lazy")
register_extra_lazy_repository_format(__name__ + '.repository',
                                      'GitRepositoryFormat')

from ..branch import network_format_registry as branch_network_format_registry

branch_network_format_registry.register_lazy(
    b'git', __name__ + '.branch', 'LocalGitBranchFormat')


from ..branch import format_registry as branch_format_registry

branch_format_registry.register_extra_lazy(
    __name__ + '.branch',
    'LocalGitBranchFormat',
    )
branch_format_registry.register_extra_lazy(
    __name__ + '.remote',
    'RemoteGitBranchFormat',
    )


from ..workingtree import format_registry as workingtree_format_registry

workingtree_format_registry.register_extra_lazy(
    __name__ + '.workingtree',
    'GitWorkingTreeFormat',
    )

controldir_network_format_registry.register_lazy(
    b'git', __name__ + ".dir", "GitControlDirFormat")


from ..diff import format_registry as diff_format_registry

diff_format_registry.register_lazy(
    'git', __name__ + '.send',
    'GitDiffTree', 'Git am-style diff format')

from ..send import format_registry as send_format_registry

send_format_registry.register_lazy('git', __name__ + '.send',
                                   'send_git', 'Git am-style diff format')

from ..directory_service import directories

directories.register_lazy('github:', __name__ + '.directory',
                          'GitHubDirectory',
                          'GitHub directory.')
directories.register_lazy('git@github.com:', __name__ + '.directory',
                          'GitHubDirectory',
                          'GitHub directory.')

from ..help_topics import topic_registry

topic_registry.register_lazy(
    'git', __name__ + '.help', 'help_git', 'Using Bazaar with Git')

from ..foreign import foreign_vcs_registry

foreign_vcs_registry.register_lazy(
    "git", __name__ + ".mapping", "foreign_vcs_git", "Stupid content tracker")


def update_git_cache(repository, revid):
    """Update the git cache after a local commit."""
    if getattr(repository, "_git", None) is not None:
        return  # No need to update cache for git repositories

    if not repository.control_transport.has("git"):
        return  # No existing cache, don't bother updating
    try:
        lazy_check_versions()
    except brz_errors.DependencyNotPresent as e:
        # dulwich is probably missing. silently ignore
        trace.mutter("not updating git map for %r: %s",
                     repository, e)

    from .object_store import BazaarObjectStore
    store = BazaarObjectStore(repository)
    with store.lock_write():
        try:
            parent_revisions = set(repository.get_parent_map([revid])[revid])
        except KeyError:
            # Isn't this a bit odd - how can a revision that was just committed
            # be missing?
            return
        missing_revisions = store._missing_revisions(parent_revisions)
        if not missing_revisions:
            store._cache.idmap.start_write_group()
            try:
                # Only update if the cache was up to date previously
                store._update_sha_map_revision(revid)
            except BaseException:
                store._cache.idmap.abort_write_group()
                raise
            else:
                store._cache.idmap.commit_write_group()


def post_commit_update_cache(local_branch, master_branch, old_revno, old_revid,
                             new_revno, new_revid):
    if local_branch is not None:
        update_git_cache(local_branch.repository, new_revid)
    update_git_cache(master_branch.repository, new_revid)


def loggerhead_git_hook(branch_app, environ):
    branch = branch_app.branch
    config_stack = branch.get_config_stack()
    if not config_stack.get('git.http'):
        return None
    from .server import git_http_hook
    return git_http_hook(branch, environ['REQUEST_METHOD'],
                         environ['PATH_INFO'])


install_lazy_named_hook("breezy.branch",
                        "Branch.hooks", "post_commit",
                        post_commit_update_cache, "git cache")
install_lazy_named_hook("breezy.plugins.loggerhead.apps.branch",
                        "BranchWSGIApp.hooks", "controller",
                        loggerhead_git_hook, "git support")


from ..config import Option, bool_from_store, option_registry

option_registry.register(
    Option('git.http',
           default=None, from_unicode=bool_from_store, invalid='warning',
           help='''\
Allow fetching of Git packs over HTTP.

This enables support for fetching Git packs over HTTP in Loggerhead.
'''))


def test_suite():
    from . import tests
    return tests.test_suite()



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from email.utils import parseaddr


class UserMapper:

    def __init__(self, lines):
        """Create a user-mapper from a list of lines.

        Blank lines and comment lines (starting with #) are ignored.
        Otherwise lines are of the form:

          old-id = new-id

        Each id may be in the following forms:

          name <email>
          name

        If old-id has the value '@', then new-id is the domain to use
        when generating an email from a user-id.
        """
        self._parse(lines)

    def _parse(self, lines):
        self._user_map = {}
        self._default_domain = None
        for line in lines:
            line = line.strip()
            if len(line) == 0 or line.startswith(b'#'):
                continue
            old, new = line.split(b'=', 1)
            old = old.strip()
            new = new.strip()
            if old == b'@':
                self._default_domain = new
                continue
            # Parse each id into a name and email address
            old_name, old_email = self._parse_id(old)
            new_name, new_email = self._parse_id(new)
            # print "found user map: %s => %s" % ((old_name, old_email), (new_name, new_email))
            self._user_map[(old_name, old_email)] = (new_name, new_email)

    def _parse_id(self, id):
        if id.find(b'<') == -1:
            return id, b""
        else:
            return parseaddr(id)

    def map_name_and_email(self, name, email):
        """Map a name and an email to the preferred name and email.

        :param name: the current name
        :param email: the current email
        :result: the preferred name and email
        """
        try:
            new_name, new_email = self._user_map[(name, email)]
        except KeyError:
            new_name = name
            if self._default_domain and not email:
                new_email = b"%s@%s" % (name, self._default_domain)
            else:
                new_email = email
        return new_name, new_email



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that WorkingTree.basis_tree() yields a valid tree."""

from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestBasisTree(TestCaseWithWorkingTree):

    def test_emtpy_tree(self):
        """A working tree with no parents."""
        tree = self.make_branch_and_tree('tree')
        basis_tree = tree.basis_tree()

        with basis_tree.lock_read():
            self.assertEqual(
                [], list(basis_tree.list_files(include_root=True)))

    def test_same_tree(self):
        """Test basis_tree when working tree hasn't been modified."""
        tree = self.make_branch_and_tree('.')
        self.build_tree(['file', 'dir/', 'dir/subfile'])
        tree.add(['file', 'dir', 'dir/subfile'])
        revision_id = tree.commit('initial tree')

        basis_tree = tree.basis_tree()
        with basis_tree.lock_read():
            self.assertEqual(revision_id, basis_tree.get_revision_id())
            # list_files() may return in either dirblock or sorted order
            # TODO: jam 20070215 Should list_files have an explicit order?
            self.assertEqual(
                ['', 'dir', 'dir/subfile', 'file'],
                sorted(info[0] for info in basis_tree.list_files(True)))

    def test_altered_tree(self):
        """Test basis really is basis after working has been modified."""
        tree = self.make_branch_and_tree('.')
        self.build_tree(['file', 'dir/', 'dir/subfile'])
        tree.add(['file', 'dir', 'dir/subfile'])
        revision_id = tree.commit('initial tree')

        self.build_tree(['new file', 'new dir/'])
        tree.rename_one('file', 'dir/new file')
        tree.unversion(['dir/subfile'])
        tree.add(['new file', 'new dir'])

        basis_tree = tree.basis_tree()
        with basis_tree.lock_read():
            self.assertEqual(revision_id, basis_tree.get_revision_id())
            # list_files() may return in either dirblock or sorted order
            # TODO: jam 20070215 Should list_files have an explicit order?
            self.assertEqual(
                ['', 'dir', 'dir/subfile', 'file'],
                sorted(info[0] for info in basis_tree.list_files(True)))



# Copyright (C) 2007, 2008, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import calendar
import re
import time

from . import osutils


def format_highres_date(t, offset=0):
    """Format a date, such that it includes higher precision in the
    seconds field.

    :param t:   The local time in fractional seconds since the epoch
    :type t: float
    :param offset:  The timezone offset in integer seconds
    :type offset: int

    Example: format_highres_date(time.time(), -time.timezone)
    this will return a date stamp for right now,
    formatted for the local timezone.

    >>> from breezy.osutils import format_date
    >>> format_date(1120153132.350850105, 0)
    'Thu 2005-06-30 17:38:52 +0000'
    >>> format_highres_date(1120153132.350850105, 0)
    'Thu 2005-06-30 17:38:52.350850105 +0000'
    >>> format_date(1120153132.350850105, -5*3600)
    'Thu 2005-06-30 12:38:52 -0500'
    >>> format_highres_date(1120153132.350850105, -5*3600)
    'Thu 2005-06-30 12:38:52.350850105 -0500'
    >>> format_highres_date(1120153132.350850105, 7200)
    'Thu 2005-06-30 19:38:52.350850105 +0200'
    >>> format_highres_date(1152428738.867522, 19800)
    'Sun 2006-07-09 12:35:38.867522001 +0530'
    """
    if not isinstance(t, float):
        raise ValueError(t)

    # This has to be formatted for "original" date, so that the
    # revision XML entry will be reproduced faithfully.
    if offset is None:
        offset = 0
    tt = time.gmtime(t + offset)

    return (osutils.weekdays[tt[6]] + time.strftime(" %Y-%m-%d %H:%M:%S", tt) +
            # Get the high-res seconds, but ignore the 0
            ('%.9f' % (t - int(t)))[1:] +
            ' %+03d%02d' % (offset / 3600, (offset / 60) % 60))


def unpack_highres_date(date):
    """This takes the high-resolution date stamp, and
    converts it back into the tuple (timestamp, timezone)
    Where timestamp is in real UTC since epoch seconds, and timezone is an
    integer number of seconds offset.

    :param date: A date formated by format_highres_date
    :type date: string

    """
    # Weekday parsing is locale sensitive, so drop the weekday
    space_loc = date.find(' ')
    if space_loc == -1 or date[:space_loc] not in osutils.weekdays:
        raise ValueError(
            'Date string does not contain a day of week: %r' % date)
    # Up until the first period is a datestamp that is generated
    # as normal from time.strftime, so use time.strptime to
    # parse it
    dot_loc = date.find('.')
    if dot_loc == -1:
        raise ValueError(
            'Date string does not contain high-precision seconds: %r' % date)
    base_time = time.strptime(date[space_loc:dot_loc], " %Y-%m-%d %H:%M:%S")
    fract_seconds, offset = date[dot_loc:].split()
    fract_seconds = float(fract_seconds)

    offset = int(offset)

    hours = int(offset / 100)
    minutes = (offset % 100)
    seconds_offset = (hours * 3600) + (minutes * 60)

    # time.mktime returns localtime, but calendar.timegm returns UTC time
    timestamp = calendar.timegm(base_time)
    timestamp -= seconds_offset
    # Add back in the fractional seconds
    timestamp += fract_seconds
    return (timestamp, seconds_offset)


def format_patch_date(secs, offset=0):
    """Format a POSIX timestamp and optional offset as a patch-style date.

    Inverse of parse_patch_date.
    """
    if offset % 60 != 0:
        raise ValueError(
            "can't represent timezone %s offset by fractional minutes" % offset)
    # so that we don't need to do calculations on pre-epoch times,
    # which doesn't work with win32 python gmtime, we always
    # give the epoch in utc
    if secs == 0:
        offset = 0
    if secs + offset < 0:
        from warnings import warn
        warn("gmtime of negative time (%s, %s) may not work on Windows" %
             (secs, offset))
    return osutils.format_date(secs, offset=offset,
                               date_fmt='%Y-%m-%d %H:%M:%S')


# Format for patch dates: %Y-%m-%d %H:%M:%S [+-]%H%M
# Groups: 1 = %Y-%m-%d %H:%M:%S; 2 = [+-]%H; 3 = %M
RE_PATCHDATE = re.compile(
    "(\\d+-\\d+-\\d+\\s+\\d+:\\d+:\\d+)\\s*([+-]\\d\\d)(\\d\\d)$")
RE_PATCHDATE_NOOFFSET = re.compile("\\d+-\\d+-\\d+\\s+\\d+:\\d+:\\d+$")


def parse_patch_date(date_str):
    """Parse a patch-style date into a POSIX timestamp and offset.

    Inverse of format_patch_date.
    """
    match = RE_PATCHDATE.match(date_str)
    if match is None:
        if RE_PATCHDATE_NOOFFSET.match(date_str) is not None:
            raise ValueError("time data %r is missing a timezone offset"
                             % date_str)
        else:
            raise ValueError("time data %r does not match format " % date_str +
                             "'%Y-%m-%d %H:%M:%S %z'")
    secs_str = match.group(1)
    offset_hours, offset_mins = int(match.group(2)), int(match.group(3))
    if abs(offset_hours) >= 24 or offset_mins >= 60:
        raise ValueError("invalid timezone %r" %
                         (match.group(2) + match.group(3)))
    offset = offset_hours * 3600 + offset_mins * 60
    tm_time = time.strptime(secs_str, '%Y-%m-%d %H:%M:%S')
    # adjust seconds according to offset before converting to POSIX
    # timestamp, to avoid edge problems
    tm_time = tm_time[:5] + (tm_time[5] - offset,) + tm_time[6:]
    secs = calendar.timegm(tm_time)
    return secs, offset



# Copyright (C) 2005-2012, 2016 Canonical Ltd
# Copyright (C) 2017-2018 Breezy developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for plugins"""

import importlib
import logging
import os
import sys
import types
from io import StringIO
from typing import Any, Dict, List

import breezy

from .. import osutils, plugin, tests

# TODO: Write a test for plugin decoration of commands.

invalidate_caches = getattr(importlib, "invalidate_caches", lambda: None)


class BaseTestPlugins(tests.TestCaseInTempDir):
    """TestCase that isolates plugin imports and cleans up on completion."""

    def setUp(self):
        super().setUp()
        self.module_name = "breezy.testingplugins"
        self.module_prefix = self.module_name + "."
        self.module = types.ModuleType(self.module_name)

        self.overrideAttr(plugin, "_MODULE_PREFIX", self.module_prefix)
        self.overrideAttr(breezy, "testingplugins", self.module)

        sys.modules[self.module_name] = self.module
        self.addCleanup(self._unregister_all)
        self.addCleanup(self._unregister_finder)

        invalidate_caches()

    def reset(self):
        """Remove all global testing state and clean up module."""
        # GZ 2017-06-02: Ideally don't do this, write new test or generate
        # bytecode by other mechanism.
        self.log("resetting plugin testing context")
        self._unregister_all()
        self._unregister_finder()
        sys.modules[self.module_name] = self.module
        for name in list(self.module.__dict__):
            if name[:2] != '__':
                delattr(self.module, name)
        invalidate_caches()
        self.plugins = None

    def update_module_paths(self, paths):
        paths = plugin.extend_path(paths, self.module_name)
        self.module.__path__ = paths
        self.log("using %r", paths)
        return paths

    def load_with_paths(self, paths, warn_load_problems=True):
        self.log("loading plugins!")
        plugin.load_plugins(
            self.update_module_paths(paths), state=self,
            warn_load_problems=warn_load_problems)

    def create_plugin(self, name, source=None, dir='.', file_name=None):
        if source is None:
            source = '''\
"""This is the doc for %s"""
''' % (name)
        if file_name is None:
            file_name = name + '.py'
        # 'source' must not fail to load
        path = osutils.pathjoin(dir, file_name)
        with open(path, 'w') as f:
            f.write(source + '\n')

    def create_plugin_package(self, name, dir=None, source=None):
        if dir is None:
            dir = name
        if source is None:
            source = '''\
"""This is the doc for {}"""
dir_source = '{}'
'''.format(name, dir)
        os.makedirs(dir)
        self.create_plugin(name, source, dir,
                           file_name='__init__.py')

    def promote_cache(self, directory):
        """Move bytecode files out of __pycache__ in given directory."""
        cache_dir = os.path.join(directory, '__pycache__')
        if os.path.isdir(cache_dir):
            for name in os.listdir(cache_dir):
                magicless_name = '.'.join(name.split('.')[0::name.count('.')])
                rel = osutils.relpath(self.test_dir, cache_dir)
                self.log("moving %s in %s to %s", name, rel, magicless_name)
                os.rename(os.path.join(cache_dir, name),
                          os.path.join(directory, magicless_name))

    def _unregister_finder(self):
        """Removes any test copies of _PluginsAtFinder from sys.meta_path."""
        idx = len(sys.meta_path)
        while idx:
            idx -= 1
            finder = sys.meta_path[idx]
            if getattr(finder, "prefix", "") == self.module_prefix:
                self.log("removed %r from sys.meta_path", finder)
                sys.meta_path.pop(idx)

    def _unregister_all(self):
        """Remove all plugins in the test namespace from sys.modules."""
        for name in list(sys.modules):
            if name.startswith(self.module_prefix) or name == self.module_name:
                self.log("removed %s from sys.modules", name)
                del sys.modules[name]

    def assertPluginModules(self, plugin_dict):
        self.assertEqual(
            {k[len(self.module_prefix):]: sys.modules[k]
                 for k in sys.modules if k.startswith(self.module_prefix)},
            plugin_dict)

    def assertPluginUnknown(self, name):
        self.assertTrue(getattr(self.module, name, None) is None)
        self.assertFalse(self.module_prefix + name in sys.modules)

    def assertPluginKnown(self, name):
        self.assertTrue(
            getattr(self.module, name, None) is not None,
            'plugins known: %r' % dir(self.module))
        self.assertTrue(self.module_prefix + name in sys.modules)


class TestLoadingPlugins(BaseTestPlugins):

    activeattributes: Dict[str, List[Any]] = {}

    def test_plugins_with_the_same_name_are_not_loaded(self):
        # This test tests that having two plugins in different directories does
        # not result in both being loaded when they have the same name.  get a
        # file name we can use which is also a valid attribute for accessing in
        # activeattributes. - we cannot give import parameters.
        tempattribute = "0"
        self.assertFalse(tempattribute in self.activeattributes)
        # set a place for the plugins to record their loading, and at the same
        # time validate that the location the plugins should record to is
        # valid and correct.
        self.__class__.activeattributes[tempattribute] = []
        self.assertTrue(tempattribute in self.activeattributes)
        # create two plugin directories
        os.mkdir('first')
        os.mkdir('second')
        # write a plugin that will record when its loaded in the
        # tempattribute list.
        template = ("from breezy.tests.test_plugins import TestLoadingPlugins\n"
                    "TestLoadingPlugins.activeattributes[%r].append('%s')\n")

        with open(os.path.join('first', 'plugin.py'), 'w') as outfile:
            outfile.write(template % (tempattribute, 'first'))
            outfile.write('\n')

        with open(os.path.join('second', 'plugin.py'), 'w') as outfile:
            outfile.write(template % (tempattribute, 'second'))
            outfile.write('\n')

        try:
            self.load_with_paths(['first', 'second'])
            self.assertEqual(['first'], self.activeattributes[tempattribute])
        finally:
            del self.activeattributes[tempattribute]

    def test_plugins_from_different_dirs_can_demand_load(self):
        self.assertFalse('breezy.plugins.pluginone' in sys.modules)
        self.assertFalse('breezy.plugins.plugintwo' in sys.modules)
        # This test tests that having two plugins in different
        # directories with different names allows them both to be loaded, when
        # we do a direct import statement.
        # Determine a file name we can use which is also a valid attribute
        # for accessing in activeattributes. - we cannot give import parameters.
        tempattribute = "different-dirs"
        self.assertFalse(tempattribute in self.activeattributes)
        # set a place for the plugins to record their loading, and at the same
        # time validate that the location the plugins should record to is
        # valid and correct.
        breezy.tests.test_plugins.TestLoadingPlugins.activeattributes[tempattribute] = [
            ]
        self.assertTrue(tempattribute in self.activeattributes)
        # create two plugin directories
        os.mkdir('first')
        os.mkdir('second')
        # write plugins that will record when they are loaded in the
        # tempattribute list.
        template = ("from breezy.tests.test_plugins import TestLoadingPlugins\n"
                    "TestLoadingPlugins.activeattributes[%r].append('%s')\n")

        with open(os.path.join('first', 'pluginone.py'), 'w') as outfile:
            outfile.write(template % (tempattribute, 'first'))
            outfile.write('\n')

        with open(os.path.join('second', 'plugintwo.py'), 'w') as outfile:
            outfile.write(template % (tempattribute, 'second'))
            outfile.write('\n')

        try:
            self.assertPluginUnknown('pluginone')
            self.assertPluginUnknown('plugintwo')
            self.update_module_paths(['first', 'second'])
            exec("import %spluginone" % self.module_prefix)
            self.assertEqual(['first'], self.activeattributes[tempattribute])
            exec("import %splugintwo" % self.module_prefix)
            self.assertEqual(['first', 'second'],
                             self.activeattributes[tempattribute])
        finally:
            del self.activeattributes[tempattribute]

    def test_plugins_can_load_from_directory_with_trailing_slash(self):
        # This test tests that a plugin can load from a directory when the
        # directory in the path has a trailing slash.
        # check the plugin is not loaded already
        self.assertPluginUnknown('ts_plugin')
        tempattribute = "trailing-slash"
        self.assertFalse(tempattribute in self.activeattributes)
        # set a place for the plugin to record its loading, and at the same
        # time validate that the location the plugin should record to is
        # valid and correct.
        breezy.tests.test_plugins.TestLoadingPlugins.activeattributes[tempattribute] = [
            ]
        self.assertTrue(tempattribute in self.activeattributes)
        # create a directory for the plugin
        os.mkdir('plugin_test')
        # write a plugin that will record when its loaded in the
        # tempattribute list.
        template = ("from breezy.tests.test_plugins import TestLoadingPlugins\n"
                    "TestLoadingPlugins.activeattributes[%r].append('%s')\n")

        with open(os.path.join('plugin_test', 'ts_plugin.py'), 'w') as outfile:
            outfile.write(template % (tempattribute, 'plugin'))
            outfile.write('\n')

        try:
            self.load_with_paths(['plugin_test' + os.sep])
            self.assertEqual(['plugin'], self.activeattributes[tempattribute])
            self.assertPluginKnown('ts_plugin')
        finally:
            del self.activeattributes[tempattribute]

    def load_and_capture(self, name, warn_load_problems=True):
        """Load plugins from '.' capturing the output.

        :param name: The name of the plugin.
        :return: A string with the log from the plugin loading call.
        """
        # Capture output
        stream = StringIO()
        try:
            handler = logging.StreamHandler(stream)
            log = logging.getLogger('brz')
            log.addHandler(handler)
            try:
                self.load_with_paths(
                    ['.'], warn_load_problems=warn_load_problems)
            finally:
                # Stop capturing output
                handler.flush()
                handler.close()
                log.removeHandler(handler)
            return stream.getvalue()
        finally:
            stream.close()

    def test_plugin_with_bad_api_version_reports(self):
        """Try loading a plugin that requests an unsupported api.

        Observe that it records the problem but doesn't complain on stderr
        when warn_load_problems=False
        """
        name = 'wants100.py'
        with open(name, 'w') as f:
            f.write("import breezy\n"
                    "from breezy.errors import IncompatibleVersion\n"
                    "raise IncompatibleVersion(breezy, [(1, 0, 0)], (0, 0, 5))\n")
        log = self.load_and_capture(name, warn_load_problems=False)
        self.assertNotContainsRe(log, r"It supports breezy version")
        self.assertEqual({'wants100'}, self.plugin_warnings.keys())
        self.assertContainsRe(
            self.plugin_warnings['wants100'][0],
            r"It supports breezy version")

    def test_plugin_with_bad_name_does_not_load(self):
        # The file name here invalid for a python module.
        name = 'brz-bad plugin-name..py'
        open(name, 'w').close()
        log = self.load_and_capture(name)
        self.assertContainsRe(
            log,
            r"Unable to load 'brz-bad plugin-name\.' in '.*' as a plugin "
            "because the file path isn't a valid module name; try renaming "
            "it to 'bad_plugin_name_'\\.")

    def test_plugin_with_error_suppress(self):
        # The file name here invalid for a python module.
        name = 'some_error.py'
        with open(name, 'w') as f:
            f.write('raise Exception("bad")\n')
        log = self.load_and_capture(name, warn_load_problems=False)
        self.assertEqual('', log)

    def test_plugin_with_error(self):
        # The file name here invalid for a python module.
        name = 'some_error.py'
        with open(name, 'w') as f:
            f.write('raise Exception("bad")\n')
        log = self.load_and_capture(name, warn_load_problems=True)
        self.assertContainsRe(
            log,
            'Unable to load plugin \'some_error\' from \'.*\': bad\n')


class TestPlugins(BaseTestPlugins):

    def setup_plugin(self, source=""):
        # This test tests a new plugin appears in breezy.plugin.plugins().
        # check the plugin is not loaded already
        self.assertPluginUnknown('plugin')
        # write a plugin that _cannot_ fail to load.
        with open('plugin.py', 'w') as f:
            f.write(source + '\n')
        self.load_with_paths(['.'])

    def test_plugin_loaded(self):
        self.assertPluginUnknown('plugin')
        self.assertIs(None, breezy.plugin.get_loaded_plugin('plugin'))
        self.setup_plugin()
        p = breezy.plugin.get_loaded_plugin('plugin')
        self.assertIsInstance(p, breezy.plugin.PlugIn)
        self.assertIs(p.module, sys.modules[self.module_prefix + 'plugin'])

    def test_plugin_loaded_disabled(self):
        self.assertPluginUnknown('plugin')
        self.overrideEnv('BRZ_DISABLE_PLUGINS', 'plugin')
        self.setup_plugin()
        self.assertIs(None, breezy.plugin.get_loaded_plugin('plugin'))

    def test_plugin_appears_in_plugins(self):
        self.setup_plugin()
        self.assertPluginKnown('plugin')
        p = self.plugins['plugin']
        self.assertIsInstance(p, breezy.plugin.PlugIn)
        self.assertIs(p.module, sys.modules[self.module_prefix + 'plugin'])

    def test_trivial_plugin_get_path(self):
        self.setup_plugin()
        p = self.plugins['plugin']
        plugin_path = self.test_dir + '/plugin.py'
        self.assertIsSameRealPath(plugin_path, osutils.normpath(p.path()))

    def test_plugin_get_path_py_not_pyc(self):
        # first import creates plugin.pyc
        self.setup_plugin()
        self.promote_cache(self.test_dir)
        self.reset()
        self.load_with_paths(['.'])  # import plugin.pyc
        p = plugin.plugins()['plugin']
        plugin_path = self.test_dir + '/plugin.py'
        self.assertIsSameRealPath(plugin_path, osutils.normpath(p.path()))

    def test_plugin_get_path_pyc_only(self):
        # first import creates plugin.pyc (or plugin.pyo depending on __debug__)
        self.setup_plugin()
        os.unlink(self.test_dir + '/plugin.py')
        self.promote_cache(self.test_dir)
        self.reset()
        self.load_with_paths(['.'])  # import plugin.pyc (or .pyo)
        p = plugin.plugins()['plugin']
        plugin_path = self.test_dir + '/plugin' + plugin.COMPILED_EXT
        self.assertIsSameRealPath(plugin_path, osutils.normpath(p.path()))

    def test_no_test_suite_gives_None_for_test_suite(self):
        self.setup_plugin()
        p = plugin.plugins()['plugin']
        self.assertEqual(None, p.test_suite())

    def test_test_suite_gives_test_suite_result(self):
        source = """def test_suite(): return 'foo'"""
        self.setup_plugin(source)
        p = plugin.plugins()['plugin']
        self.assertEqual('foo', p.test_suite())

    def test_no_load_plugin_tests_gives_None_for_load_plugin_tests(self):
        self.setup_plugin()
        loader = tests.TestUtil.TestLoader()
        p = plugin.plugins()['plugin']
        self.assertEqual(None, p.load_plugin_tests(loader))

    def test_load_plugin_tests_gives_load_plugin_tests_result(self):
        source = """
def load_tests(loader, standard_tests, pattern):
    return 'foo'"""
        self.setup_plugin(source)
        loader = tests.TestUtil.TestLoader()
        p = plugin.plugins()['plugin']
        self.assertEqual('foo', p.load_plugin_tests(loader))

    def check_version_info(self, expected, source='', name='plugin'):
        self.setup_plugin(source)
        self.assertEqual(expected, plugin.plugins()[name].version_info())

    def test_no_version_info(self):
        self.check_version_info(None)

    def test_with_version_info(self):
        self.check_version_info((1, 2, 3, 'dev', 4),
                                "version_info = (1, 2, 3, 'dev', 4)")

    def test_short_version_info_gets_padded(self):
        # the gtk plugin has version_info = (1,2,3) rather than the 5-tuple.
        # so we adapt it
        self.check_version_info((1, 2, 3, 'final', 0),
                                "version_info = (1, 2, 3)")

    def check_version(self, expected, source=None, name='plugin'):
        self.setup_plugin(source)
        self.assertEqual(expected, plugins[name].__version__)

    def test_no_version_info___version__(self):
        self.setup_plugin()
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("unknown", plugin.__version__)

    def test_str__version__with_version_info(self):
        self.setup_plugin("version_info = '1.2.3'")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3", plugin.__version__)

    def test_noniterable__version__with_version_info(self):
        self.setup_plugin("version_info = (1)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1", plugin.__version__)

    def test_1__version__with_version_info(self):
        self.setup_plugin("version_info = (1,)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1", plugin.__version__)

    def test_1_2__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2", plugin.__version__)

    def test_1_2_3__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3", plugin.__version__)

    def test_candidate__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3, 'candidate', 1)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3.rc1", plugin.__version__)

    def test_dev__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3, 'dev', 0)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3.dev", plugin.__version__)

    def test_dev_fallback__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3, 'dev', 4)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3.dev4", plugin.__version__)

    def test_final__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3, 'final', 0)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3", plugin.__version__)

    def test_final_fallback__version__with_version_info(self):
        self.setup_plugin("version_info = (1, 2, 3, 'final', 2)")
        plugin = breezy.plugin.plugins()['plugin']
        self.assertEqual("1.2.3.2", plugin.__version__)


class TestHelpIndex(tests.TestCase):
    """Tests for the PluginsHelpIndex class."""

    def test_default_constructable(self):
        index = plugin.PluginsHelpIndex()

    def test_get_topics_None(self):
        """Searching for None returns an empty list."""
        index = plugin.PluginsHelpIndex()
        self.assertEqual([], index.get_topics(None))

    def test_get_topics_for_plugin(self):
        """Searching for plugin name gets its docstring."""
        index = plugin.PluginsHelpIndex()
        # make a new plugin here for this test, even if we're run with
        # --no-plugins
        self.assertFalse('breezy.plugins.demo_module' in sys.modules)
        demo_module = FakeModule('', 'breezy.plugins.demo_module')
        sys.modules['breezy.plugins.demo_module'] = demo_module
        try:
            topics = index.get_topics('demo_module')
            self.assertEqual(1, len(topics))
            self.assertIsInstance(topics[0], plugin.ModuleHelpTopic)
            self.assertEqual(demo_module, topics[0].module)
        finally:
            del sys.modules['breezy.plugins.demo_module']

    def test_get_topics_no_topic(self):
        """Searching for something that is not a plugin returns []."""
        # test this by using a name that cannot be a plugin - its not
        # a valid python identifier.
        index = plugin.PluginsHelpIndex()
        self.assertEqual([], index.get_topics('nothing by this name'))

    def test_prefix(self):
        """PluginsHelpIndex has a prefix of 'plugins/'."""
        index = plugin.PluginsHelpIndex()
        self.assertEqual('plugins/', index.prefix)

    def test_get_plugin_topic_with_prefix(self):
        """Searching for plugins/demo_module returns help."""
        index = plugin.PluginsHelpIndex()
        self.assertFalse('breezy.plugins.demo_module' in sys.modules)
        demo_module = FakeModule('', 'breezy.plugins.demo_module')
        sys.modules['breezy.plugins.demo_module'] = demo_module
        try:
            topics = index.get_topics('plugins/demo_module')
            self.assertEqual(1, len(topics))
            self.assertIsInstance(topics[0], plugin.ModuleHelpTopic)
            self.assertEqual(demo_module, topics[0].module)
        finally:
            del sys.modules['breezy.plugins.demo_module']


class FakeModule:
    """A fake module to test with."""

    def __init__(self, doc, name):
        self.__doc__ = doc
        self.__name__ = name


class TestModuleHelpTopic(tests.TestCase):
    """Tests for the ModuleHelpTopic class."""

    def test_contruct(self):
        """Construction takes the module to document."""
        mod = FakeModule('foo', 'foo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual(mod, topic.module)

    def test_get_help_text_None(self):
        """A ModuleHelpTopic returns the docstring for get_help_text."""
        mod = FakeModule(None, 'demo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual("Plugin 'demo' has no docstring.\n",
                         topic.get_help_text())

    def test_get_help_text_no_carriage_return(self):
        """ModuleHelpTopic.get_help_text adds a \n if needed."""
        mod = FakeModule('one line of help', 'demo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual("one line of help\n",
                         topic.get_help_text())

    def test_get_help_text_carriage_return(self):
        """ModuleHelpTopic.get_help_text adds a \n if needed."""
        mod = FakeModule('two lines of help\nand more\n', 'demo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual("two lines of help\nand more\n",
                         topic.get_help_text())

    def test_get_help_text_with_additional_see_also(self):
        mod = FakeModule('two lines of help\nand more', 'demo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual(
            "two lines of help\nand more\n\n:See also: bar, foo\n",
            topic.get_help_text(['foo', 'bar']))

    def test_get_help_topic(self):
        """The help topic for a plugin is its module name."""
        mod = FakeModule('two lines of help\nand more', 'breezy.plugins.demo')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual('demo', topic.get_help_topic())
        mod = FakeModule('two lines of help\nand more',
                         'breezy.plugins.foo_bar')
        topic = plugin.ModuleHelpTopic(mod)
        self.assertEqual('foo_bar', topic.get_help_topic())


class TestEnvPluginPath(tests.TestCase):

    user = "USER"
    core = "CORE"
    site = "SITE"

    def check_path(self, expected_dirs, setting_dirs):
        if setting_dirs is None:
            del os.environ['BRZ_PLUGIN_PATH']
        else:
            os.environ['BRZ_PLUGIN_PATH'] = os.pathsep.join(setting_dirs)
        actual = [(p if t == 'path' else t.upper())
                  for p, t in plugin._env_plugin_path()]
        self.assertEqual(expected_dirs, actual)

    def test_default(self):
        self.check_path([self.user, self.core, self.site], None)

    def test_adhoc_policy(self):
        self.check_path([self.user, self.core, self.site],
                        ['+user', '+core', '+site'])

    def test_fallback_policy(self):
        self.check_path([self.core, self.site, self.user],
                        ['+core', '+site', '+user'])

    def test_override_policy(self):
        self.check_path([self.user, self.site, self.core],
                        ['+user', '+site', '+core'])

    def test_disable_user(self):
        self.check_path([self.core, self.site], ['-user'])

    def test_disable_user_twice(self):
        # Ensures multiple removals don't left cruft
        self.check_path([self.core, self.site], ['-user', '-user'])

    def test_duplicates_are_removed(self):
        self.check_path([self.user, self.core, self.site],
                        ['+user', '+user'])
        # And only the first reference is kept (since the later references will
        # only produce '<plugin> already loaded' mutters)
        self.check_path([self.user, self.core, self.site],
                        ['+user', '+user', '+core',
                         '+user', '+site', '+site',
                         '+core'])

    def test_disable_overrides_enable(self):
        self.check_path([self.core, self.site], ['-user', '+user'])

    def test_disable_core(self):
        self.check_path([self.site], ['-core'])
        self.check_path([self.user, self.site], ['+user', '-core'])

    def test_disable_site(self):
        self.check_path([self.core], ['-site'])
        self.check_path([self.user, self.core], ['-site', '+user'])

    def test_override_site(self):
        self.check_path(['mysite', self.user, self.core],
                        ['mysite', '-site', '+user'])
        self.check_path(['mysite', self.core],
                        ['mysite', '-site'])

    def test_override_core(self):
        self.check_path(['mycore', self.user, self.site],
                        ['mycore', '-core', '+user', '+site'])
        self.check_path(['mycore', self.site],
                        ['mycore', '-core'])

    def test_my_plugin_only(self):
        self.check_path(['myplugin'], ['myplugin', '-user', '-core', '-site'])

    def test_my_plugin_first(self):
        self.check_path(['myplugin', self.core, self.site, self.user],
                        ['myplugin', '+core', '+site', '+user'])

    def test_bogus_references(self):
        self.check_path(['+foo', '-bar', self.core, self.site],
                        ['+foo', '-bar'])


class TestDisablePlugin(BaseTestPlugins):

    def test_cannot_import(self):
        self.create_plugin_package('works')
        self.create_plugin_package('fails')
        self.overrideEnv('BRZ_DISABLE_PLUGINS', 'fails')
        self.update_module_paths(["."])
        import breezy.testingplugins.works as works
        try:
            import breezy.testingplugins.fails as fails
        except ImportError:
            pass
        else:
            self.fail("Loaded blocked plugin: " + repr(fails))
        self.assertPluginModules({'fails': None, 'works': works})

    def test_partial_imports(self):
        self.create_plugin('good')
        self.create_plugin('bad')
        self.create_plugin_package('ugly')
        self.overrideEnv('BRZ_DISABLE_PLUGINS', 'bad:ugly')
        self.load_with_paths(['.'])
        self.assertEqual({'good'}, self.plugins.keys())
        self.assertPluginModules({
            'good': self.plugins['good'].module,
            'bad': None,
            'ugly': None,
        })
        # Ensure there are no warnings about plugins not being imported as
        # the user has explictly requested they be disabled.
        self.assertNotContainsRe(self.get_log(), r"Unable to load plugin")


class TestEnvDisablePlugins(tests.TestCase):

    def _get_names(self, env_value):
        os.environ['BRZ_DISABLE_PLUGINS'] = env_value
        return plugin._env_disable_plugins()

    def test_unset(self):
        self.assertEqual([], plugin._env_disable_plugins())

    def test_empty(self):
        self.assertEqual([], self._get_names(''))

    def test_single(self):
        self.assertEqual(['single'], self._get_names('single'))

    def test_multi(self):
        expected = ['one', 'two']
        self.assertEqual(expected, self._get_names(os.pathsep.join(expected)))

    def test_mixed(self):
        value = os.pathsep.join(['valid', 'in-valid'])
        self.assertEqual(['valid'], self._get_names(value))
        self.assertContainsRe(
            self.get_log(),
            r"Invalid name 'in-valid' in BRZ_DISABLE_PLUGINS=" + repr(value))


class TestEnvPluginsAt(tests.TestCase):

    def _get_paths(self, env_value):
        os.environ['BRZ_PLUGINS_AT'] = env_value
        return plugin._env_plugins_at()

    def test_empty(self):
        self.assertEqual([], plugin._env_plugins_at())
        self.assertEqual([], self._get_paths(''))

    def test_one_path(self):
        self.assertEqual([('b', os.path.abspath('man'))], self._get_paths('b@man'))

    def test_multiple(self):
        self.assertEqual(
            [('tools', os.path.abspath('bzr-tools')),
             ('p', os.path.abspath('play.py'))],
            self._get_paths(os.pathsep.join(('tools@bzr-tools', 'p@play.py'))))

    def test_many_at(self):
        self.assertEqual(
            [('church', os.path.abspath('StMichael@Plea@Norwich'))],
            self._get_paths('church@StMichael@Plea@Norwich'))

    def test_only_py(self):
        self.assertEqual(
            [('test', os.path.abspath('test.py'))],
            self._get_paths('./test.py'))

    def test_only_package(self):
        self.assertEqual([('py', '/opt/b/py')], self._get_paths('/opt/b/py'))

    def test_bad_name(self):
        self.assertEqual([], self._get_paths('/usr/local/bzr-git'))
        self.assertContainsRe(
            self.get_log(),
            r"Invalid name 'bzr-git' in BRZ_PLUGINS_AT='/usr/local/bzr-git'")


class TestLoadPluginAt(BaseTestPlugins):

    def setUp(self):
        super().setUp()
        # Create the same plugin in two directories
        self.create_plugin_package('test_foo', dir='non-standard-dir')
        # The "normal" directory, we use 'standard' instead of 'plugins' to
        # avoid depending on the precise naming.
        self.create_plugin_package('test_foo', dir='standard/test_foo')

    def assertTestFooLoadedFrom(self, path):
        self.assertPluginKnown('test_foo')
        self.assertDocstring('This is the doc for test_foo',
                             self.module.test_foo)
        self.assertEqual(path, self.module.test_foo.dir_source)

    def test_regular_load(self):
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom('standard/test_foo')

    def test_import(self):
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        self.update_module_paths(['standard'])
        import breezy.testingplugins.test_foo
        self.assertTestFooLoadedFrom('non-standard-dir')

    def test_loading(self):
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom('non-standard-dir')

    def test_loading_other_name(self):
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        os.rename('standard/test_foo', 'standard/test_bar')
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom('non-standard-dir')

    def test_compiled_loaded(self):
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom('non-standard-dir')
        self.assertIsSameRealPath('non-standard-dir/__init__.py',
                                  self.module.test_foo.__file__)

        # Try importing again now that the source has been compiled
        os.remove('non-standard-dir/__init__.py')
        self.promote_cache('non-standard-dir')
        self.reset()
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom('non-standard-dir')
        suffix = plugin.COMPILED_EXT
        self.assertIsSameRealPath('non-standard-dir/__init__' + suffix,
                                  self.module.test_foo.__file__)

    def test_submodule_loading(self):
        # We create an additional directory under the one for test_foo
        self.create_plugin_package('test_bar', dir='non-standard-dir/test_bar')
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        self.update_module_paths(['standard'])
        import breezy.testingplugins.test_foo  # noqa: F401
        self.assertEqual(self.module_prefix + 'test_foo',
                         self.module.test_foo.__package__)
        import breezy.testingplugins.test_foo.test_bar  # noqa: F401
        self.assertIsSameRealPath('non-standard-dir/test_bar/__init__.py',
                                  self.module.test_foo.test_bar.__file__)

    def test_relative_submodule_loading(self):
        self.create_plugin_package('test_foo', dir='another-dir', source='''
from . import test_bar
''')
        # We create an additional directory under the one for test_foo
        self.create_plugin_package('test_bar', dir='another-dir/test_bar')
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@another-dir')
        self.update_module_paths(['standard'])
        import breezy.testingplugins.test_foo  # noqa: F401
        self.assertEqual(self.module_prefix + 'test_foo',
                         self.module.test_foo.__package__)
        self.assertIsSameRealPath('another-dir/test_bar/__init__.py',
                                  self.module.test_foo.test_bar.__file__)

    def test_loading_from___init__only(self):
        # We rename the existing __init__.py file to ensure that we don't load
        # a random file
        init = 'non-standard-dir/__init__.py'
        random = 'non-standard-dir/setup.py'
        os.rename(init, random)
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@non-standard-dir')
        self.load_with_paths(['standard'])
        self.assertPluginUnknown('test_foo')

    def test_loading_from_specific_file(self):
        plugin_dir = 'non-standard-dir'
        plugin_file_name = 'iamtestfoo.py'
        plugin_path = osutils.pathjoin(plugin_dir, plugin_file_name)
        source = '''\
"""This is the doc for {}"""
dir_source = '{}'
'''.format('test_foo', plugin_path)
        self.create_plugin('test_foo', source=source,
                           dir=plugin_dir, file_name=plugin_file_name)
        self.overrideEnv('BRZ_PLUGINS_AT', 'test_foo@%s' % plugin_path)
        self.load_with_paths(['standard'])
        self.assertTestFooLoadedFrom(plugin_path)


class TestDescribePlugins(BaseTestPlugins):

    def test_describe_plugins(self):
        class DummyModule:
            __doc__ = 'Hi there'

        class DummyPlugin:
            __version__ = '0.1.0'
            module = DummyModule()
        self.plugin_warnings = {'bad': ['Failed to load (just testing)']}
        self.plugins = {'good': DummyPlugin()}
        self.assertEqual("""\
bad (failed to load)
  ** Failed to load (just testing)

good 0.1.0
  Hi there

""", ''.join(plugin.describe_plugins(state=self)))



# Copyright (C) 2005, 2006, 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Export a Tree to a zip file.
"""

import os
import stat
import sys
import tempfile
import time
import zipfile
from contextlib import closing

from .. import osutils
from ..export import _export_iter_entries
from ..trace import mutter

# Windows expects this bit to be set in the 'external_attr' section,
# or it won't consider the entry a directory.
ZIP_DIRECTORY_BIT = (1 << 4)
FILE_PERMISSIONS = (0o644 << 16)
DIR_PERMISSIONS = (0o755 << 16)

_FILE_ATTR = stat.S_IFREG | FILE_PERMISSIONS
_DIR_ATTR = stat.S_IFDIR | ZIP_DIRECTORY_BIT | DIR_PERMISSIONS


def zip_archive_generator(tree, dest, root, subdir=None,
                          force_mtime=None, recurse_nested=False):
    """ Export this tree to a new zip file.

    `dest` will be created holding the contents of this tree; if it
    already exists, it will be overwritten".
    """
    compression = zipfile.ZIP_DEFLATED
    with tempfile.SpooledTemporaryFile() as buf:
        with closing(zipfile.ZipFile(buf, "w", compression)) as zipf, \
                tree.lock_read():
            for dp, tp, ie in _export_iter_entries(
                    tree, subdir, recurse_nested=recurse_nested):
                mutter("  export {%s} kind %s to %s", tp, ie.kind, dest)

                # zipfile.ZipFile switches all paths to forward
                # slashes anyway, so just stick with that.
                if force_mtime is not None:
                    mtime = force_mtime
                else:
                    mtime = tree.get_file_mtime(tp)
                date_time = time.localtime(mtime)[:6]
                filename = osutils.pathjoin(root, dp)
                if ie.kind == "file":
                    zinfo = zipfile.ZipInfo(
                        filename=filename,
                        date_time=date_time)
                    zinfo.compress_type = compression
                    zinfo.external_attr = _FILE_ATTR
                    content = tree.get_file_text(tp)
                    zipf.writestr(zinfo, content)
                elif ie.kind in ("directory", "tree-reference"):
                    # Directories must contain a trailing slash, to indicate
                    # to the zip routine that they are really directories and
                    # not just empty files.
                    zinfo = zipfile.ZipInfo(
                        filename=filename + '/',
                        date_time=date_time)
                    zinfo.compress_type = compression
                    zinfo.external_attr = _DIR_ATTR
                    zipf.writestr(zinfo, '')
                elif ie.kind == "symlink":
                    zinfo = zipfile.ZipInfo(
                        filename=(filename + '.lnk'),
                        date_time=date_time)
                    zinfo.compress_type = compression
                    zinfo.external_attr = _FILE_ATTR
                    zipf.writestr(zinfo, tree.get_symlink_target(tp))
        # Urgh, headers are written last since they include e.g. file size.
        # So we have to buffer it all :(
        buf.seek(0)
        yield from osutils.file_iterator(buf)



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Transport decorator that logs transport operations to brz.log."""

# see also the transportstats plugin, which gives you some summary information
# in a machine-readable dump

import time
import types

from ..trace import mutter
from ..transport import decorator


class TransportLogDecorator(decorator.TransportDecorator):
    """Decorator for Transports that logs interesting operations to brz.log.

    In general we want to log things that usually take a network round trip
    and may be slow.

    Not all operations are logged yet.

    See also TransportTraceDecorator, that records a machine-readable log in
    memory for eg testing.
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        def _make_hook(hookname):
            def _hook(relpath, *args, **kw):
                return self._log_and_call(hookname, relpath, *args, **kw)
            return _hook
        # GZ 2017-05-21: Not all methods take relpath as first argument, for
        # instance copy_to takes list of relpaths. Also, unclear on url vs
        # filesystem path split. Needs tidying up.
        for methodname in (
                'append_bytes',
                'append_file',
                'copy_to',
                'delete',
                'get',
                'has',
                'open_write_stream',
                'mkdir',
                'move',
                'put_bytes', 'put_bytes_non_atomic', 'put_file put_file_non_atomic',
                'list_dir', 'lock_read', 'lock_write',
                'readv', 'rename', 'rmdir',
                'stat',
                'ulock',
                ):
            setattr(self, methodname, _make_hook(methodname))

    @classmethod
    def _get_url_prefix(self):
        return 'log+'

    def iter_files_recursive(self):
        # needs special handling because it does not have a relpath parameter
        mutter("%s %s"
               % ('iter_files_recursive', self._decorated.base))
        return self._call_and_log_result('iter_files_recursive', (), {})

    def _log_and_call(self, methodname, relpath, *args, **kwargs):
        if kwargs:
            kwargs_str = dict(kwargs)
        else:
            kwargs_str = ''
        mutter("%s %s %s %s"
               % (methodname, relpath,
                  self._shorten(self._strip_tuple_parens(args)),
                  kwargs_str))
        return self._call_and_log_result(methodname, (relpath,) + args, kwargs)

    def _call_and_log_result(self, methodname, args, kwargs):
        before = time.time()
        try:
            result = getattr(self._decorated, methodname)(*args, **kwargs)
        except Exception as e:
            mutter("  --> %s" % e)
            mutter("      %.03fs" % (time.time() - before))
            raise
        return self._show_result(before, methodname, result)

    def _show_result(self, before, methodname, result):
        result_len = None
        if isinstance(result, types.GeneratorType):
            # We now consume everything from the generator so that we can show
            # the results and the time it took to get them.  However, to keep
            # compatibility with callers that may specifically expect a result
            # (see <https://launchpad.net/bugs/340347>) we also return a new
            # generator, reset to the starting position.
            result = list(result)
            return_result = iter(result)
        else:
            return_result = result
        # Is this an io object with a getvalue() method?
        getvalue = getattr(result, 'getvalue', None)
        if getvalue is not None:
            val = repr(getvalue())
            result_len = len(val)
            shown_result = "%s(%s) (%d bytes)" % (result.__class__.__name__,
                                                  self._shorten(val), result_len)
        elif methodname == 'readv':
            num_hunks = len(result)
            total_bytes = sum((len(d) for o, d in result))
            shown_result = "readv response, %d hunks, %d total bytes" % (
                num_hunks, total_bytes)
            result_len = total_bytes
        else:
            shown_result = self._shorten(self._strip_tuple_parens(result))
        mutter("  --> %s" % shown_result)
        # The log decorator no longer shows the elapsed time or transfer rate
        # because they're available in the log prefixes and the transport
        # activity display respectively.
        if False:
            elapsed = time.time() - before
            if result_len and elapsed > 0:
                # this is the rate of higher-level data, not the raw network
                # speed using base-10 units (see HACKING.txt).
                mutter("      %9.03fs %8dkB/s"
                       % (elapsed, result_len / elapsed / 1000))
            else:
                mutter("      %9.03fs" % (elapsed))
        return return_result

    def _shorten(self, x):
        if len(x) > 70:
            x = x[:67] + '...'
        return x

    def _strip_tuple_parens(self, t):
        t = repr(t)
        if t[0] == '(' and t[-1] == ')':
            t = t[1:-1]
        return t


def get_test_permutations():
    """Return the permutations to be used in testing."""
    from ..tests import test_server
    return [(TransportLogDecorator, test_server.LogDecoratorServer)]



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""UI implementations for use in testing.
"""


from .. import progress, ui


class ProgressRecordingUIFactory(ui.UIFactory, progress.DummyProgress):
    """Captures progress updates made through it.

    This is overloaded as both the UIFactory and the progress model."""

    def __init__(self):
        super().__init__()
        self._calls = []
        self.depth = 0

    def nested_progress_bar(self):
        self.depth += 1
        return self

    def finished(self):
        self.depth -= 1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finished()
        return False

    def update(self, message, count=None, total=None):
        if self.depth == 1:
            self._calls.append(("update", count, total, message))
